# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""A {{glossary("microgrid")}} is a local electrical grid that connects a set of
electrical components together.  They are often built around a passive power consumer,
to supplement the electricity consumed from the {{glossary("grid", "public grid")}} with
on-site power generation or storage systems.

Microgrids can also function in {{glossary("island", "island-mode")}}, without a grid
connection, or without a local power consumer, but they have to have at least one of the
two, to be meaningful.

## Frequenz SDK Microgrid Model

The SDK aims to provide an abstract model of the microgrid that enables high-level
interactions with {{glossary("component", "microgrid components")}}, without having to
worry about (or even be aware of) location-specific details such as:

- where the {{glossary("meter", "meters")}} are placed,
- how many {{glossary("battery", "batteries")}},
- whether there's a grid connection or a passive consumer,
- what models the {{glossary("inverter", "inverters")}} are, etc.
- whether components are having downtimes, because {{glossary("metric", "metrics")}} and
  limits get adjusted automatically when components are having downtimes.

Users of the SDK can develop applications around this interface once and deploy
anywhere, and the SDK will take care of translating the requests and instructions to
correspond to the specific microgrid configurations.

``` mermaid
flowchart LR

subgraph Left[Measurements only]
direction LR
  grid["Grid Connection"]
  consumer["Consumer"]
  pv["PV Arrays"]
  chp["CHP"]
end

junction(( ))

subgraph Right[Measurements and control]
direction LR
  bat["Batteries"]
  ev["EV Chargers"]
end

grid --- junction
consumer --- junction
pv --- junction
chp --- junction

junction --- bat
junction --- ev
```

## Grid

This refers to a microgrid's connection to the external Grid.  The power flowing through
this connection can be streamed through
[`grid_power`][frequenz.sdk.timeseries.grid.Grid.power].

In locations without a grid connection, this method remains accessible, and streams zero
values.

## Consumer

This is the main power consumer at the site of a microgrid, and often the
{{glossary("load")}} the microgrid is built to support.  The power drawn by the consumer
is available through [`consumer_power`][frequenz.sdk.timeseries.consumer.Consumer.power]

In locations without a consumer, this method streams zero values.

## Producers: PV Arrays, CHP

The total CHP production in a site can be streamed through
[`chp_power`][frequenz.sdk.timeseries.logical_meter.LogicalMeter.chp_power].  PV Power
is available through the PV pool described below.  And total producer power is available
through [`microgrid.producer().power`][frequenz.sdk.timeseries.producer.Producer.power].

As is the case with the other methods, if PV Arrays or CHPs are not available in a
microgrid, the corresponding methods stream zero values.

## PV Arrays

The total PV power production is available through
[`pv_pool`][frequenz.sdk.microgrid.new_pv_pool]'s
[`power`][frequenz.sdk.timeseries.pv_pool.PVPool.power].  The PV pool by default uses
all PV inverters available at a location, but PV pool instances can be created for
subsets of PV inverters if necessary, by specifying the inverter ids.

The `pv_pool` also provides available power bounds through the
[`power_status`][frequenz.sdk.timeseries.pv_pool.PVPool.power_status] method.

The `pv_pool` also provides a control method
[`propose_power`][frequenz.sdk.timeseries.pv_pool.PVPool.propose_power], which accepts
values in the {{glossary("psc", "Passive Sign Convention")}} and supports only
production.


## Batteries

The total Battery power is available through the
[`battery_pool`][frequenz.sdk.microgrid.new_battery_pool]'s
[`power`][frequenz.sdk.timeseries.battery_pool.BatteryPool.power].  The battery pool by
default uses all batteries available at a location, but battery pool instances can be
created for subsets of batteries if necessary, by specifying the battery ids.

The `battery_pool` also provides
[`soc`][frequenz.sdk.timeseries.battery_pool.BatteryPool.soc],
[`capacity`][frequenz.sdk.timeseries.battery_pool.BatteryPool.capacity],
[`temperature`][frequenz.sdk.timeseries.battery_pool.BatteryPool.temperature] and
available power bounds through the
[`power_status`][frequenz.sdk.timeseries.battery_pool.BatteryPool.power_status] method.

The `battery_pool` also provides control methods
[`propose_power`][frequenz.sdk.timeseries.battery_pool.BatteryPool.propose_power] (which
accepts values in the {{glossary("psc", "Passive Sign Convention")}} and supports both
charging and discharging), or through
[`propose_charge`][frequenz.sdk.timeseries.battery_pool.BatteryPool.propose_charge], or
[`propose_discharge`][frequenz.sdk.timeseries.battery_pool.BatteryPool.propose_discharge].

## EV Chargers

The [`ev_charger_pool`][frequenz.sdk.microgrid.new_ev_charger_pool] offers a
[`power`][frequenz.sdk.timeseries.ev_charger_pool.EVChargerPool.power] method that
streams the total power measured for all the {{glossary("ev-charger", "EV Chargers")}}
at a site.

The `ev_charger_pool` also provides available power bounds through the
[`power_status`][frequenz.sdk.timeseries.ev_charger_pool.EVChargerPool.power_status]
method.


The `ev_charger_pool` also provides a control method
[`propose_power`][frequenz.sdk.timeseries.ev_charger_pool.EVChargerPool.propose_power],
which accepts values in the {{glossary("psc", "Passive Sign Convention")}} and supports
only charging.

# Component pools

The SDK provides a unified interface for interacting with sets of Batteries, EV
chargers and PV arrays, through their corresponding `Pool`s.

* [Battery pool][frequenz.sdk.microgrid.new_battery_pool]
* [EV charger pool][frequenz.sdk.microgrid.new_ev_charger_pool]
* [PV pool][frequenz.sdk.microgrid.new_pv_pool]

All of them provide support for streaming aggregated data and for setting the
power values of the components.

## Streaming component data

All pools have a `power` property, which is a
[`FormulaEngine`][frequenz.sdk.timeseries.formula_engine.FormulaEngine] that can

- provide a stream of resampled power values, which correspond to the sum of the
power measured from all the components in the pool together.

- be composed with other power streams to for composite formulas.

In addition, the battery pool has some additional properties that can be used as
streams for metrics specific to batteries:
[`soc`][frequenz.sdk.timeseries.battery_pool.BatteryPool.soc],
[`capacity`][frequenz.sdk.timeseries.battery_pool.BatteryPool.capacity] and
[`temperature`][frequenz.sdk.timeseries.battery_pool.BatteryPool.temperature].

## Setting power

All pools provide a `propose_power` method for setting power for the pool.  This
would then be distributed to the individual components in the pool, using an
algorithm that's suitable for the category of the components.  For example, when
controlling batteries, power could be distributed based on the `SoC` of the
individual batteries, to keep the batteries in balance.

### Resolving conflicting power proposals

When there are multiple actors trying to control the same set of batteries, a
target power is calculated based on the priorities of the actors making the
requests.  Actors need to specify their priorities as parameters when creating
the `*Pool` instances using the constructors mentioned above.

The algorithm used for resolving power conflicts based on actor priority can be
found in the documentation for any of the
[`propose_power`][frequenz.sdk.timeseries.battery_pool.BatteryPool.propose_power]
methods.

### Shifting the target power by an Operating Point power

There are cases where the target power needs to be shifted by an operating point.  This
can be done by designating some actors to be able to set only the operating point power.

When creating a `*Pool` instance using the above-mentioned constructors, an optional
`set_operating_point` parameter can be passed to specify that this actor is special, and
the target power of the regular actors will be shifted by the target power of all actors
with `set_operating_point` together.

In a location with 2 regular actors and 1 `set_operating_point` actor, here's how things
would play out:

1. When only regular actors have made proposals, the power bounds available from the
   batteries are available to them exactly.

   | actor priority | in op group? | proposed power/bounds | available bounds |
   |----------------|--------------|-----------------------|------------------|
   | 3              | No           | 1000, -4000..2500     | -3000..3000      |
   | 2              | No           | 2500                  | -3000..2500      |
   | 1              | Yes          | None                  | -3000..3000      |

   Power actually distributed to the batteries: 2500W

2. When the `set_operating_point` actor has made proposals, the bounds available to the
   regular actors gets shifted, and the final power that actually gets distributed to
   the batteries is also shifted.

   | actor priority | in op group? | proposed power/bounds | available bounds |
   |----------------|--------------|-----------------------|------------------|
   | 3              | No           | 1000, -4000..2500     | -2000..4000      |
   | 2              | No           | 2500                  | -2000..2500      |
   | 1              | Yes          | -1000                 | -3000..3000      |

   Power actually distributed to the batteries: 1500W
"""  # noqa: D205, D400

from datetime import timedelta

from ..actor import ResamplerConfig
from . import _data_pipeline, connection_manager
from ._data_pipeline import (
    consumer,
    frequency,
    grid,
    logical_meter,
    new_battery_pool,
    new_ev_charger_pool,
    new_pv_pool,
    producer,
    voltage_per_phase,
)


async def initialize(
    server_url: str,
    resampler_config: ResamplerConfig,
    *,
    api_power_request_timeout: timedelta = timedelta(seconds=5.0),
) -> None:
    """Initialize the microgrid connection manager and the data pipeline.

    Args:
        server_url: The location of the microgrid API server in the form of a URL.
            The following format is expected: `grpc://hostname{:port}{?ssl=ssl}`,
            where the port should be an int between `0` and `65535` (defaulting to
            `9090`) and ssl should be a boolean (defaulting to false). For example:
            `grpc://localhost:1090?ssl=true`.
        resampler_config: Configuration for the resampling actor.
        api_power_request_timeout: Timeout to use when making power requests to
            the microgrid API.  When requests to components timeout, they will
            be marked as blocked for a short duration, during which time they
            will be unavailable from the corresponding component pools.
    """
    await connection_manager.initialize(server_url)
    await _data_pipeline.initialize(
        resampler_config,
        api_power_request_timeout=api_power_request_timeout,
    )


__all__ = [
    "initialize",
    "consumer",
    "grid",
    "frequency",
    "logical_meter",
    "new_battery_pool",
    "new_ev_charger_pool",
    "new_pv_pool",
    "producer",
    "voltage_per_phase",
]



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""The Microgrid API data source for the DataSourcingActor."""

import asyncio
import logging
from collections.abc import Callable
from typing import Any

from frequenz.channels import Receiver, Sender
from frequenz.client.microgrid import (
    BatteryData,
    ComponentCategory,
    ComponentMetricId,
    EVChargerData,
    InverterData,
    MeterData,
)
from frequenz.quantities import Quantity

from ..._internal._asyncio import run_forever
from ..._internal._channels import ChannelRegistry
from ...microgrid import connection_manager
from ...timeseries import Sample
from ._component_metric_request import ComponentMetricRequest

_logger = logging.getLogger(__name__)

_MeterDataMethods: dict[ComponentMetricId, Callable[[MeterData], float]] = {
    ComponentMetricId.ACTIVE_POWER: lambda msg: msg.active_power,
    ComponentMetricId.ACTIVE_POWER_PHASE_1: lambda msg: msg.active_power_per_phase[0],
    ComponentMetricId.ACTIVE_POWER_PHASE_2: lambda msg: msg.active_power_per_phase[1],
    ComponentMetricId.ACTIVE_POWER_PHASE_3: lambda msg: msg.active_power_per_phase[2],
    ComponentMetricId.CURRENT_PHASE_1: lambda msg: msg.current_per_phase[0],
    ComponentMetricId.CURRENT_PHASE_2: lambda msg: msg.current_per_phase[1],
    ComponentMetricId.CURRENT_PHASE_3: lambda msg: msg.current_per_phase[2],
    ComponentMetricId.VOLTAGE_PHASE_1: lambda msg: msg.voltage_per_phase[0],
    ComponentMetricId.VOLTAGE_PHASE_2: lambda msg: msg.voltage_per_phase[1],
    ComponentMetricId.VOLTAGE_PHASE_3: lambda msg: msg.voltage_per_phase[2],
    ComponentMetricId.FREQUENCY: lambda msg: msg.frequency,
    ComponentMetricId.REACTIVE_POWER: lambda msg: msg.reactive_power,
    ComponentMetricId.REACTIVE_POWER_PHASE_1: lambda msg: msg.reactive_power_per_phase[
        0
    ],
    ComponentMetricId.REACTIVE_POWER_PHASE_2: lambda msg: msg.reactive_power_per_phase[
        1
    ],
    ComponentMetricId.REACTIVE_POWER_PHASE_3: lambda msg: msg.reactive_power_per_phase[
        2
    ],
}

_BatteryDataMethods: dict[ComponentMetricId, Callable[[BatteryData], float]] = {
    ComponentMetricId.SOC: lambda msg: msg.soc,
    ComponentMetricId.SOC_LOWER_BOUND: lambda msg: msg.soc_lower_bound,
    ComponentMetricId.SOC_UPPER_BOUND: lambda msg: msg.soc_upper_bound,
    ComponentMetricId.CAPACITY: lambda msg: msg.capacity,
    ComponentMetricId.POWER_INCLUSION_LOWER_BOUND: lambda msg: (
        msg.power_inclusion_lower_bound
    ),
    ComponentMetricId.POWER_EXCLUSION_LOWER_BOUND: lambda msg: (
        msg.power_exclusion_lower_bound
    ),
    ComponentMetricId.POWER_EXCLUSION_UPPER_BOUND: lambda msg: (
        msg.power_exclusion_upper_bound
    ),
    ComponentMetricId.POWER_INCLUSION_UPPER_BOUND: lambda msg: (
        msg.power_inclusion_upper_bound
    ),
    ComponentMetricId.TEMPERATURE: lambda msg: msg.temperature,
}

_InverterDataMethods: dict[ComponentMetricId, Callable[[InverterData], float]] = {
    ComponentMetricId.ACTIVE_POWER: lambda msg: msg.active_power,
    ComponentMetricId.ACTIVE_POWER_PHASE_1: lambda msg: msg.active_power_per_phase[0],
    ComponentMetricId.ACTIVE_POWER_PHASE_2: lambda msg: msg.active_power_per_phase[1],
    ComponentMetricId.ACTIVE_POWER_PHASE_3: lambda msg: msg.active_power_per_phase[2],
    ComponentMetricId.ACTIVE_POWER_INCLUSION_LOWER_BOUND: lambda msg: (
        msg.active_power_inclusion_lower_bound
    ),
    ComponentMetricId.ACTIVE_POWER_EXCLUSION_LOWER_BOUND: lambda msg: (
        msg.active_power_exclusion_lower_bound
    ),
    ComponentMetricId.ACTIVE_POWER_EXCLUSION_UPPER_BOUND: lambda msg: (
        msg.active_power_exclusion_upper_bound
    ),
    ComponentMetricId.ACTIVE_POWER_INCLUSION_UPPER_BOUND: lambda msg: (
        msg.active_power_inclusion_upper_bound
    ),
    ComponentMetricId.CURRENT_PHASE_1: lambda msg: msg.current_per_phase[0],
    ComponentMetricId.CURRENT_PHASE_2: lambda msg: msg.current_per_phase[1],
    ComponentMetricId.CURRENT_PHASE_3: lambda msg: msg.current_per_phase[2],
    ComponentMetricId.VOLTAGE_PHASE_1: lambda msg: msg.voltage_per_phase[0],
    ComponentMetricId.VOLTAGE_PHASE_2: lambda msg: msg.voltage_per_phase[1],
    ComponentMetricId.VOLTAGE_PHASE_3: lambda msg: msg.voltage_per_phase[2],
    ComponentMetricId.FREQUENCY: lambda msg: msg.frequency,
    ComponentMetricId.REACTIVE_POWER: lambda msg: msg.reactive_power,
    ComponentMetricId.REACTIVE_POWER_PHASE_1: lambda msg: msg.reactive_power_per_phase[
        0
    ],
    ComponentMetricId.REACTIVE_POWER_PHASE_2: lambda msg: msg.reactive_power_per_phase[
        1
    ],
    ComponentMetricId.REACTIVE_POWER_PHASE_3: lambda msg: msg.reactive_power_per_phase[
        2
    ],
}

_EVChargerDataMethods: dict[ComponentMetricId, Callable[[EVChargerData], float]] = {
    ComponentMetricId.ACTIVE_POWER: lambda msg: msg.active_power,
    ComponentMetricId.ACTIVE_POWER_PHASE_1: lambda msg: msg.active_power_per_phase[0],
    ComponentMetricId.ACTIVE_POWER_PHASE_2: lambda msg: msg.active_power_per_phase[1],
    ComponentMetricId.ACTIVE_POWER_PHASE_3: lambda msg: msg.active_power_per_phase[2],
    ComponentMetricId.CURRENT_PHASE_1: lambda msg: msg.current_per_phase[0],
    ComponentMetricId.CURRENT_PHASE_2: lambda msg: msg.current_per_phase[1],
    ComponentMetricId.CURRENT_PHASE_3: lambda msg: msg.current_per_phase[2],
    ComponentMetricId.VOLTAGE_PHASE_1: lambda msg: msg.voltage_per_phase[0],
    ComponentMetricId.VOLTAGE_PHASE_2: lambda msg: msg.voltage_per_phase[1],
    ComponentMetricId.VOLTAGE_PHASE_3: lambda msg: msg.voltage_per_phase[2],
    ComponentMetricId.FREQUENCY: lambda msg: msg.frequency,
    ComponentMetricId.REACTIVE_POWER: lambda msg: msg.reactive_power,
    ComponentMetricId.REACTIVE_POWER_PHASE_1: lambda msg: msg.reactive_power_per_phase[
        0
    ],
    ComponentMetricId.REACTIVE_POWER_PHASE_2: lambda msg: msg.reactive_power_per_phase[
        1
    ],
    ComponentMetricId.REACTIVE_POWER_PHASE_3: lambda msg: msg.reactive_power_per_phase[
        2
    ],
}


class MicrogridApiSource:
    """Fetches requested metrics from the Microgrid API.

    Used by the DataSourcingActor.
    """

    def __init__(
        self,
        registry: ChannelRegistry,
    ) -> None:
        """Create a `MicrogridApiSource` instance.

        Args:
            registry: A channel registry.  To be replaced by a singleton
                instance.
        """
        self._comp_categories_cache: dict[int, ComponentCategory] = {}

        self.comp_data_receivers: dict[int, Receiver[Any]] = {}
        """The dictionary of component IDs to data receivers."""

        self.comp_data_tasks: dict[int, asyncio.Task[None]] = {}
        """The dictionary of component IDs to asyncio tasks."""

        self._registry = registry
        self._req_streaming_metrics: dict[
            int, dict[ComponentMetricId, list[ComponentMetricRequest]]
        ] = {}

    async def _get_component_category(self, comp_id: int) -> ComponentCategory | None:
        """Get the component category of the given component.

        Args:
            comp_id: Id of the requested component.

        Returns:
            The category of the given component, if it is a valid component, or None
                otherwise.
        """
        if comp_id in self._comp_categories_cache:
            return self._comp_categories_cache[comp_id]

        api = connection_manager.get().api_client
        for comp in await api.components():
            self._comp_categories_cache[comp.component_id] = comp.category

        if comp_id in self._comp_categories_cache:
            return self._comp_categories_cache[comp_id]

        return None

    async def _check_battery_request(
        self,
        comp_id: int,
        requests: dict[ComponentMetricId, list[ComponentMetricRequest]],
    ) -> None:
        """Check if the requests are valid Battery metrics.

        Raises:
            ValueError: if the requested metric is not available for batteries.

        Args:
            comp_id: The id of the requested component.
            requests: A list of metric requests received from external actors
                for the given battery.
        """
        for metric in requests:
            if metric not in _BatteryDataMethods:
                err = f"Unknown metric {metric} for Battery id {comp_id}"
                _logger.error(err)
                raise ValueError(err)
        if comp_id not in self.comp_data_receivers:
            self.comp_data_receivers[comp_id] = (
                await connection_manager.get().api_client.battery_data(comp_id)
            )

    async def _check_ev_charger_request(
        self,
        comp_id: int,
        requests: dict[ComponentMetricId, list[ComponentMetricRequest]],
    ) -> None:
        """Check if the requests are valid EV Charger metrics.

        Raises:
            ValueError: if the requested metric is not available for ev charger.

        Args:
            comp_id: The id of the requested component.
            requests: A list of metric requests received from external actors
                for the given EV Charger.
        """
        for metric in requests:
            if metric not in _EVChargerDataMethods:
                err = f"Unknown metric {metric} for EvCharger id {comp_id}"
                _logger.error(err)
                raise ValueError(err)
        if comp_id not in self.comp_data_receivers:
            self.comp_data_receivers[comp_id] = (
                await connection_manager.get().api_client.ev_charger_data(comp_id)
            )

    async def _check_inverter_request(
        self,
        comp_id: int,
        requests: dict[ComponentMetricId, list[ComponentMetricRequest]],
    ) -> None:
        """Check if the requests are valid Inverter metrics.

        Raises:
            ValueError: if the requested metric is not available for inverters.

        Args:
            comp_id: The id of the requested component.
            requests: A list of metric requests received from external actors
                for the given inverter.
        """
        for metric in requests:
            if metric not in _InverterDataMethods:
                err = f"Unknown metric {metric} for Inverter id {comp_id}"
                _logger.error(err)
                raise ValueError(err)
        if comp_id not in self.comp_data_receivers:
            self.comp_data_receivers[comp_id] = (
                await connection_manager.get().api_client.inverter_data(comp_id)
            )

    async def _check_meter_request(
        self,
        comp_id: int,
        requests: dict[ComponentMetricId, list[ComponentMetricRequest]],
    ) -> None:
        """Check if the requests are valid Meter metrics.

        Raises:
            ValueError: if the requested metric is not available for meters.

        Args:
            comp_id: The id of the requested component.
            requests: A list of metric requests received from external actors
                for the given meter.
        """
        for metric in requests:
            if metric not in _MeterDataMethods:
                err = f"Unknown metric {metric} for Meter id {comp_id}"
                _logger.error(err)
                raise ValueError(err)
        if comp_id not in self.comp_data_receivers:
            self.comp_data_receivers[comp_id] = (
                await connection_manager.get().api_client.meter_data(comp_id)
            )

    async def _check_requested_component_and_metrics(
        self,
        comp_id: int,
        category: ComponentCategory,
        requests: dict[ComponentMetricId, list[ComponentMetricRequest]],
    ) -> None:
        """Check if the requested component and metrics are valid.

        Raises:
            ValueError: if the category is unknown or if the requested metric
                is unavailable to the given category.

        Args:
            comp_id: The id of the requested component.
            category: The category of the requested component.
            requests: A list of metric requests received from external actors
                for the given component.
        """
        if comp_id in self.comp_data_receivers:
            return

        if category == ComponentCategory.BATTERY:
            await self._check_battery_request(comp_id, requests)
        elif category == ComponentCategory.EV_CHARGER:
            await self._check_ev_charger_request(comp_id, requests)
        elif category == ComponentCategory.INVERTER:
            await self._check_inverter_request(comp_id, requests)
        elif category == ComponentCategory.METER:
            await self._check_meter_request(comp_id, requests)
        else:
            err = f"Unknown component category {category}"
            _logger.error(err)
            raise ValueError(err)

    def _get_data_extraction_method(
        self, category: ComponentCategory, metric: ComponentMetricId
    ) -> Callable[[Any], float]:
        """Get the data extraction method for the given metric.

        Raises:
            ValueError: if the category is unknown.

        Args:
            category: The category of the component.
            metric: The metric for which we need an extraction method.

        Returns:
            A method that accepts a `ComponentData` object and returns a float
                representing the given metric.
        """
        if category == ComponentCategory.BATTERY:
            return _BatteryDataMethods[metric]
        if category == ComponentCategory.INVERTER:
            return _InverterDataMethods[metric]
        if category == ComponentCategory.METER:
            return _MeterDataMethods[metric]
        if category == ComponentCategory.EV_CHARGER:
            return _EVChargerDataMethods[metric]
        err = f"Unknown component category {category}"
        _logger.error(err)
        raise ValueError(err)

    def _get_metric_senders(
        self,
        category: ComponentCategory,
        requests: dict[ComponentMetricId, list[ComponentMetricRequest]],
    ) -> list[tuple[Callable[[Any], float], list[Sender[Sample[Quantity]]]]]:
        """Get channel senders from the channel registry for each requested metric.

        Args:
            category: The category of the component.
            requests: A list of metric requests received from external actors for a
                certain component.

        Returns:
            A dictionary of output metric names to channel senders from the channel
                registry.
        """
        return [
            (
                self._get_data_extraction_method(category, metric),
                [
                    self._registry.get_or_create(
                        Sample[Quantity], request.get_channel_name()
                    ).new_sender()
                    for request in req_list
                ],
            )
            for (metric, req_list) in requests.items()
        ]

    async def _handle_data_stream(
        self,
        comp_id: int,
        category: ComponentCategory,
    ) -> None:
        """Stream component data and send the requested metrics out.

        Args:
            comp_id: Id of the requested component.
            category: The category of the component.

        Raises:
            Exception: if an error occurs while handling the data stream.
        """
        try:
            stream_senders = []
            if comp_id in self._req_streaming_metrics:
                await self._check_requested_component_and_metrics(
                    comp_id, category, self._req_streaming_metrics[comp_id]
                )
                stream_senders = self._get_metric_senders(
                    category, self._req_streaming_metrics[comp_id]
                )
            api_data_receiver: Receiver[Any] = self.comp_data_receivers[comp_id]

            async def process_msg(data: Any) -> None:
                async with asyncio.TaskGroup() as tg:
                    for extractor, senders in stream_senders:
                        for sender in senders:
                            sample = Sample(data.timestamp, Quantity(extractor(data)))
                            name = f"send:ts={sample.timestamp}:cid={comp_id}"
                            tg.create_task(sender.send(sample), name=name)

            sending_tasks: set[asyncio.Task[None]] = set()

            async def clean_tasks(
                sending_tasks: set[asyncio.Task[None]],
            ) -> set[asyncio.Task[None]]:
                done, pending = await asyncio.wait(sending_tasks, timeout=0)
                for task in done:
                    try:
                        task.result()
                    # pylint: disable-next=broad-except
                    except (asyncio.CancelledError, Exception):
                        _logger.exception(
                            "Error while processing message in task %s",
                            task.get_name(),
                        )
                return pending

            async for data in api_data_receiver:
                name = f"process_msg:cid={comp_id}"
                sending_tasks.add(asyncio.create_task(process_msg(data), name=name))
                sending_tasks = await clean_tasks(sending_tasks)

            await asyncio.gather(*sending_tasks)
            await asyncio.gather(
                *[
                    self._registry.close_and_remove(r.get_channel_name())
                    for requests in self._req_streaming_metrics[comp_id].values()
                    for r in requests
                ]
            )
        except Exception:
            _logger.exception(
                "Unexpected error while handling data stream for component %d (%s), "
                "component data is not being streamed anymore",
                comp_id,
                category.name,
            )
            raise

    async def _update_streams(
        self,
        comp_id: int,
        category: ComponentCategory,
    ) -> None:
        """Update the requested metric streams for the given component.

        Args:
            comp_id: Id of the requested component.
            category: Category of the requested component.
        """
        if comp_id in self.comp_data_tasks:
            self.comp_data_tasks[comp_id].cancel()

        self.comp_data_tasks[comp_id] = asyncio.create_task(
            run_forever(lambda: self._handle_data_stream(comp_id, category)),
            name=f"{type(self).__name__}._update_stream({comp_id=}, {category.name})",
        )

    async def add_metric(self, request: ComponentMetricRequest) -> None:
        """Add a metric to be streamed from the microgrid API.

        Args:
            request: A request object for a metric, received from a downstream
                actor.
        """
        comp_id = request.component_id

        category = await self._get_component_category(comp_id)

        if category is None:
            _logger.error("Unknown component ID: %d in request %s", comp_id, request)
            return

        self._req_streaming_metrics.setdefault(comp_id, {}).setdefault(
            request.metric_id, []
        )

        for existing_request in self._req_streaming_metrics[comp_id][request.metric_id]:
            if existing_request.get_channel_name() == request.get_channel_name():
                # the requested metric is already being handled, so nothing to do.
                return

        self._req_streaming_metrics[comp_id][request.metric_id].append(request)

        await self._update_streams(
            comp_id,
            category,
        )



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Configuration for the power distributor's EV charger manager."""

from collections import abc
from dataclasses import dataclass, field
from datetime import timedelta

from frequenz.quantities import Current


@dataclass(frozen=True)
class EVDistributionConfig:
    """Configuration for the power distributor's EV charger manager."""

    component_ids: abc.Set[int]
    """The component ids of the EV chargers."""

    min_current: Current = field(default_factory=lambda: Current.from_amperes(6.0))
    """The minimum current that can be allocated to an EV charger."""

    initial_current: Current = field(default_factory=lambda: Current.from_amperes(10.0))
    """The initial current that can be allocated to an EV charger."""

    increase_power_interval: timedelta = timedelta(seconds=60)
    """The interval at which the power can be increased for an EV charger."""



# License: MIT
# Copyright  © 2023 Frequenz Energy-as-a-Service GmbH

"""Formula generator from component graph for CHP Power."""


import logging
from collections import abc

from frequenz.client.microgrid import ComponentCategory, ComponentMetricId
from frequenz.quantities import Power

from ....microgrid import connection_manager
from ...formula_engine import FormulaEngine
from ._formula_generator import (
    NON_EXISTING_COMPONENT_ID,
    FormulaGenerationError,
    FormulaGenerator,
)

_logger = logging.getLogger(__name__)


class CHPPowerFormula(FormulaGenerator[Power]):
    """Formula generator for CHP Power."""

    def generate(  # noqa: DOC502 (FormulaGenerationError is raised indirectly by _get_chp_meters)
        self,
    ) -> FormulaEngine[Power]:
        """Make a formula for the cumulative CHP power of a microgrid.

        The calculation is performed by adding the active power measurements from
        dedicated meters attached to CHPs.

        Returns:
            A formula engine that will calculate cumulative CHP power values.

        Raises:
            FormulaGenerationError: If there's no dedicated meter attached to every CHP.

        """
        builder = self._get_builder(
            "chp-power", ComponentMetricId.ACTIVE_POWER, Power.from_watts
        )

        chp_meter_ids = self._get_chp_meters()
        if not chp_meter_ids:
            _logger.warning("No CHPs found in the component graph.")
            builder.push_component_metric(
                NON_EXISTING_COMPONENT_ID, nones_are_zeros=True
            )
            return builder.build()

        for idx, chp_meter_id in enumerate(chp_meter_ids):
            if idx > 0:
                builder.push_oper("+")
            builder.push_component_metric(chp_meter_id, nones_are_zeros=False)

        return builder.build()

    def _get_chp_meters(self) -> abc.Set[int]:
        """Get the meter IDs of the CHPs from the component graph.

        Returns:
            A set of meter IDs of the CHPs in the component graph. If no CHPs are
                found, None is returned.

        Raises:
            FormulaGenerationError: If there's no dedicated meter attached to every CHP.
        """
        component_graph = connection_manager.get().component_graph
        chps = list(
            comp
            for comp in component_graph.components()
            if comp.category == ComponentCategory.CHP
        )

        chp_meters: set[int] = set()
        for chp in chps:
            predecessors = component_graph.predecessors(chp.component_id)
            if len(predecessors) != 1:
                raise FormulaGenerationError(
                    f"CHP {chp.component_id} has {len(predecessors)} predecessors. "
                    " Expected exactly one."
                )
            meter = next(iter(predecessors))
            if meter.category != ComponentCategory.METER:
                raise FormulaGenerationError(
                    f"CHP {chp.component_id} has a predecessor of category "
                    f"{meter.category}. Expected ComponentCategory.METER."
                )
            meter_successors = component_graph.successors(meter.component_id)
            if not all(successor in chps for successor in meter_successors):
                raise FormulaGenerationError(
                    f"Meter {meter.component_id} connected to CHP {chp.component_id}"
                    "has non-chp successors."
                )
            chp_meters.add(meter.component_id)
        return chp_meters



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Power distribution state tracking for ev chargers."""


from dataclasses import dataclass
from datetime import datetime
from typing import Iterable

from frequenz.client.microgrid import EVChargerData
from frequenz.quantities import Power


@dataclass
class EvcState:
    """A class for tracking state of an ev charger."""

    component_id: int
    """The component id of the ev charger."""

    last_data: EVChargerData
    """The last data received from the EV charger."""

    power: Power
    """The power currently used by the EV charger."""

    last_allocation: Power
    """The last allocation made for the EV."""

    last_reallocation_time: datetime
    """The last time the ev charger was allocated power.

    Used to make sure we don't allocate power to the ev charger too often.
    """

    last_charging_time: datetime
    """The last time the ev charger was charging.

    Used to de-allocate power from the ev charger if it has not been charging
    for a while.
    """

    def update_last_allocation(self, allocation: Power, alloc_time: datetime) -> None:
        """Update the last allocation and related timestamps.

        Args:
            allocation: The most allocation allocation made for the EV.
            alloc_time: The time at which the allocation was made.
        """
        self.last_allocation = allocation
        self.last_reallocation_time = alloc_time
        self.last_charging_time = alloc_time

    def update_state(
        self,
        latest_ev_data: EVChargerData,
    ) -> None:
        """Update EvcState from component data.

        Args:
            latest_ev_data: latest ev data from component data stream.
        """
        self.power = Power.from_watts(latest_ev_data.active_power)
        self.last_data = latest_ev_data

        if self.power > Power.zero():
            self.last_charging_time = latest_ev_data.timestamp


class EvcStates:
    """Tracks states of all ev chargers."""

    _states: dict[int, EvcState]

    def __init__(self) -> None:
        """Initialize this instance."""
        self._states = {}

    def get_ev_total_used_power(self) -> Power:
        """Return the total power consumed by all EV Chargers."""
        total_used = Power.zero()
        for evc in self._states.values():
            total_used += evc.power
        return total_used

    def get_total_allocated_power(self) -> Power:
        """Return the total power allocated to all EV Chargers."""
        total_allocated = Power.zero()
        for evc in self._states.values():
            total_allocated += evc.last_allocation
        return total_allocated

    def get(self, component_id: int) -> EvcState:
        """Return a reference to the EvcState object with the given component_id.

        Args:
            component_id: identifies the object to retrieve.

        Returns:
            The EvcState object with the given component_id.
        """
        return self._states[component_id]

    def add_evc(self, state: EvcState) -> None:
        """Add the given EvcState object to the list.

        Args:
            state: The EvcState object to add to the list.
        """
        self._states[state.component_id] = state

    def values(self) -> Iterable[EvcState]:
        """Return an iterator over all EvcState objects.

        Returns:
            An iterator over all EvcState objects.
        """
        return self._states.values()

    def __contains__(self, component_id: int) -> bool:
        """Check if the given component_id has an associated EvcState object.

        Args:
            component_id: The component id to test.

        Returns:
            Boolean indicating whether the given component_id is a known
            EvCharger.
        """
        return component_id in self._states



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Base formula generator from component graph for Grid Power."""

from abc import ABC, abstractmethod

from frequenz.client.microgrid import Component, ComponentCategory

from ..._base_types import QuantityT
from .._formula_engine import FormulaEngine
from .._resampled_formula_builder import ResampledFormulaBuilder
from ._fallback_formula_metric_fetcher import FallbackFormulaMetricFetcher
from ._formula_generator import ComponentNotFound, FormulaGenerator


class GridPowerFormulaBase(FormulaGenerator[QuantityT], ABC):
    """Base class for grid power formula generators."""

    def _generate(
        self, builder: ResampledFormulaBuilder[QuantityT]
    ) -> FormulaEngine[QuantityT]:
        """Generate a formula for calculating grid power from the component graph.

        Args:
            builder: The builder to use to create the formula.

        Returns:
            A formula engine that will calculate grid power values.

        Raises:
            ComponentNotFound: when the component graph doesn't have a `GRID` component.
        """
        grid_successors = self._get_grid_component_successors()

        components = {
            c
            for c in grid_successors
            if c.category
            in {
                ComponentCategory.INVERTER,
                ComponentCategory.EV_CHARGER,
                ComponentCategory.METER,
            }
        }

        if not components:
            raise ComponentNotFound("No grid successors found")

        # generate a formula that just adds values from all components that are
        # directly connected to the grid.  If the requested formula type is
        # `PASSIVE_SIGN_CONVENTION`, there is nothing more to do.  If the requested
        # formula type is `PRODUCTION`, the formula output is negated, then clipped to
        # 0.  If the requested formula type is `CONSUMPTION`, the formula output is
        # already positive, so it is just clipped to 0.
        #
        # So the formulas would look like:
        #  - `PASSIVE_SIGN_CONVENTION`: `(grid-successor-1 + grid-successor-2 + ...)`
        #  - `PRODUCTION`: `max(0, -(grid-successor-1 + grid-successor-2 + ...))`
        #  - `CONSUMPTION`: `max(0, (grid-successor-1 + grid-successor-2 + ...))`
        if self._config.allow_fallback:
            fallbacks = self._get_fallback_formulas(components)

            for idx, (primary_component, fallback_formula) in enumerate(
                fallbacks.items()
            ):
                if idx > 0:
                    builder.push_oper("+")

                # should only be the case if the component is not a meter
                builder.push_component_metric(
                    primary_component.component_id,
                    nones_are_zeros=(
                        primary_component.category != ComponentCategory.METER
                    ),
                    fallback=fallback_formula,
                )
        else:
            for idx, comp in enumerate(components):
                if idx > 0:
                    builder.push_oper("+")

                builder.push_component_metric(
                    comp.component_id,
                    nones_are_zeros=(comp.category != ComponentCategory.METER),
                )

        return builder.build()

    @abstractmethod
    def _get_fallback_formulas(
        self, components: set[Component]
    ) -> dict[Component, FallbackFormulaMetricFetcher[QuantityT] | None]:
        """Find primary and fallback components and create fallback formulas.

        The primary component is the one that will be used to calculate the producer power.
        If it is not available, the fallback formula will be used instead.
        Fallback formulas calculate the grid power using the fallback components.
        Fallback formulas are wrapped in `FallbackFormulaMetricFetcher`.

        Args:
            components: The producer components.

        Returns:
            A dictionary mapping primary components to their FallbackFormulaMetricFetcher.
        """



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Grid connection point.

This module provides the `Grid` type, which represents a grid connection point
in a microgrid.
"""
from __future__ import annotations

import logging
import uuid
from dataclasses import dataclass

from frequenz.channels import Sender
from frequenz.client.microgrid._component import ComponentCategory, ComponentMetricId
from frequenz.quantities import Current, Power, ReactivePower

from .._internal._channels import ChannelRegistry
from ..microgrid import connection_manager
from ..microgrid._data_sourcing import ComponentMetricRequest
from ._fuse import Fuse
from .formula_engine import FormulaEngine, FormulaEngine3Phase
from .formula_engine._formula_engine_pool import FormulaEnginePool
from .formula_engine._formula_generators import (
    GridCurrentFormula,
    GridPower3PhaseFormula,
    GridPowerFormula,
    GridReactivePowerFormula,
)

_logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Grid:
    """A grid connection point.

    !!! note
        The `Grid` instance is not meant to be created directly by users.
        Use the [`microgrid.grid`][frequenz.sdk.microgrid.grid] method for
        creating or getting the `Grid` instance.

    Example:
        ```python
        from datetime import timedelta

        from frequenz.sdk import microgrid
        from frequenz.sdk.timeseries import ResamplerConfig

        await microgrid.initialize(
            "grpc://127.0.0.1:50051",
            ResamplerConfig(resampling_period=timedelta(seconds=1))
        )

        grid = microgrid.grid()

        # Get a receiver for a builtin formula
        grid_power_recv = grid.power.new_receiver()
        async for grid_power_sample in grid_power_recv:
            print(grid_power_sample)
        ```
    """

    fuse: Fuse | None
    """The fuse protecting the grid connection point.

    The rated current of the fuse is set to zero in case of an islanded
    microgrid.
    And the fuse is set to `None` when the grid connection component metadata
    lacks information about the fuse.
    """

    _formula_pool: FormulaEnginePool
    """The formula engine pool to generate grid metrics."""

    @property
    def power(self) -> FormulaEngine[Power]:
        """Fetch the grid power for the microgrid.

        This formula produces values that are in the Passive Sign Convention (PSC).

        If a formula engine to calculate grid power is not already running, it will be
        started.

        A receiver from the formula engine can be created using the `new_receiver`
        method.

        Returns:
            A FormulaEngine that will calculate and stream grid power.
        """
        engine = self._formula_pool.from_power_formula_generator(
            "grid_power",
            GridPowerFormula,
        )
        assert isinstance(engine, FormulaEngine)
        return engine

    @property
    def reactive_power(self) -> FormulaEngine[ReactivePower]:
        """Fetch the grid reactive power for the microgrid.

        This formula produces values that are in the Passive Sign Convention (PSC).

        If a formula engine to calculate grid power is not already running, it will be
        started.

        A receiver from the formula engine can be created using the `new_receiver`
        method.

        Returns:
            A FormulaEngine that will calculate and stream grid reactive power.
        """
        engine = self._formula_pool.from_reactive_power_formula_generator(
            f"grid-{ComponentMetricId.REACTIVE_POWER.value}",
            GridReactivePowerFormula,
        )
        assert isinstance(engine, FormulaEngine)
        return engine

    @property
    def _power_per_phase(self) -> FormulaEngine3Phase[Power]:
        """Fetch the per-phase grid power for the microgrid.

        This formula produces values that are in the Passive Sign Convention (PSC).

        A receiver from the formula engine can be created using the
        `new_receiver`method.

        Returns:
            A FormulaEngine that will calculate and stream grid 3-phase power.
        """
        engine = self._formula_pool.from_power_3_phase_formula_generator(
            "grid_power_3_phase", GridPower3PhaseFormula
        )
        assert isinstance(engine, FormulaEngine3Phase)
        return engine

    @property
    def current_per_phase(self) -> FormulaEngine3Phase[Current]:
        """Fetch the per-phase grid current for the microgrid.

        This formula produces values that are in the Passive Sign Convention (PSC).

        If a formula engine to calculate grid current is not already running, it will be
        started.

        A receiver from the formula engine can be created using the `new_receiver`
        method.

        Returns:
            A FormulaEngine that will calculate and stream grid current.
        """
        engine = self._formula_pool.from_3_phase_current_formula_generator(
            "grid_current",
            GridCurrentFormula,
        )
        assert isinstance(engine, FormulaEngine3Phase)
        return engine

    async def stop(self) -> None:
        """Stop all formula engines."""
        await self._formula_pool.stop()


_GRID: Grid | None = None


def initialize(
    channel_registry: ChannelRegistry,
    resampler_subscription_sender: Sender[ComponentMetricRequest],
) -> None:
    """Initialize the grid connection.

    Args:
        channel_registry: The channel registry instance shared with the
            resampling actor.
        resampler_subscription_sender: The sender for sending metric requests
            to the resampling actor.

    Raises:
        RuntimeError: If there is more than 1 grid connection point in the
            microgrid, or if the grid connection point is not initialized.
    """
    global _GRID  # pylint: disable=global-statement

    grid_connections = list(
        connection_manager.get().component_graph.components(
            component_categories={ComponentCategory.GRID},
        )
    )

    grid_connections_count = len(grid_connections)

    fuse: Fuse | None = None

    match grid_connections_count:
        case 0:
            fuse = Fuse(max_current=Current.zero())
            _logger.info(
                "No grid connection found for this microgrid. "
                "This is normal for an islanded microgrid."
            )
        case 1:
            metadata = grid_connections[0].metadata
            if metadata is None:
                _logger.warning(
                    "Unable to get grid metadata, the grid connection point is "
                    "considered to have no fuse"
                )
            elif metadata.fuse is None:
                _logger.warning("The grid connection point does not have a fuse")
            else:
                fuse = Fuse(max_current=Current.from_amperes(metadata.fuse.max_current))
        case _:
            raise RuntimeError(
                f"Expected at most one grid connection, got {grid_connections_count}"
            )

    namespace = f"grid-{uuid.uuid4()}"
    formula_pool = FormulaEnginePool(
        namespace,
        channel_registry,
        resampler_subscription_sender,
    )

    _GRID = Grid(fuse, formula_pool)


def get() -> Grid:
    """Get the grid connection.

    Note that the rated current of the fuse is set to zero in case of an
    islanded microgrid.
    And the fuse is set to `None` when the grid connection component metadata
    lacks information about the fuse.

    Returns:
        The grid connection.
    """
    assert _GRID, "Grid is not initialized"
    return _GRID



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Fetch and stream the phase-to-neutral voltage from a source component.

Each phase of the phase-to-neutral voltage is fetched from the source
component individually and then streamed as a 3-phase sample.
"""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING

from frequenz.channels import Receiver, Sender
from frequenz.client.microgrid import Component, ComponentCategory, ComponentMetricId
from frequenz.quantities import Quantity, Voltage

from .._internal._channels import ChannelRegistry
from ..timeseries._base_types import Sample, Sample3Phase

if TYPE_CHECKING:
    # Imported here to avoid a circular import.
    from ..microgrid._data_sourcing import ComponentMetricRequest

_logger = logging.getLogger(__name__)


class VoltageStreamer:
    """Fetch and stream the phase-to-neutral voltage from a source component.

    Example:
        ```python
        from datetime import timedelta

        from frequenz.sdk import microgrid
        from frequenz.sdk.timeseries import ResamplerConfig

        await microgrid.initialize(
            "grpc://127.0.0.1:50051",
            ResamplerConfig(resampling_period=timedelta(seconds=1))
        )

        # Get a receiver for the phase-to-neutral voltage.
        voltage_recv = microgrid.voltage_per_phase().new_receiver()

        async for voltage_sample in voltage_recv:
            print(voltage_sample)
        ```
    """

    def __init__(
        self,
        resampler_subscription_sender: Sender[ComponentMetricRequest],
        channel_registry: ChannelRegistry,
        source_component: Component | None = None,
    ):
        """Initialize the phase-to-neutral voltage streaming.

        Args:
            resampler_subscription_sender: The sender for sending metric
                requests to the resampling actor.
            channel_registry: The channel registry for the phase-to-neutral
                voltage streaming.
            source_component: The source component to receive the
                phase-to-neutral voltage. If None, it fetches the source
                component from the connection manager.

        """
        self._resampler_subscription_sender = resampler_subscription_sender
        """The sender for sending metric requests to the resampling actor."""

        self._channel_registry = channel_registry
        """The channel registry for the phase-to-neutral voltage streaming."""

        from ..microgrid import (  # pylint: disable=import-outside-toplevel
            connection_manager,
        )

        if not source_component:
            component_graph = connection_manager.get().component_graph
            source_component = component_graph.find_first_descendant_component(
                root_category=ComponentCategory.GRID,
                descendant_categories=[
                    ComponentCategory.METER,
                    ComponentCategory.INVERTER,
                    ComponentCategory.EV_CHARGER,
                ],
            )

        self._source_component = source_component
        """The source component to receive the phase-to-neutral voltage."""

        self._task: None | asyncio.Task[None] = None
        """The task for fetching and stream the phase-to-neutral voltage."""

        self._namespace = "microgrid-voltage"
        """The namespace for the phase-to-neutral voltage streaming."""

        self._channel_key = f"{self._namespace}-all-phases"
        """The channel key for the phase-to-neutral voltage streaming."""

    @property
    def source(self) -> Component:
        """Get the component to fetch the phase-to-neutral voltage from.

        Returns:
            The component to fetch the phase-to-neutral voltage from.
        """
        return self._source_component

    def new_receiver(self) -> Receiver[Sample3Phase[Voltage]]:
        """Create a receiver for the phase-to-neutral voltage.

        Each phase of the phase-to-neutral voltage is fetched from the source
        component individually and then streamed as a 3-phase sample.

        Returns:
            A receiver that will receive the phase-to-neutral voltage as a
            3-phase sample.
        """
        receiver = self._channel_registry.get_or_create(
            Sample3Phase[Voltage], self._channel_key
        ).new_receiver()

        if not self._task:
            self._task = asyncio.create_task(self._send_request())
        else:
            _logger.info("Voltage request already sent: %s", self._source_component)

        return receiver

    async def _send_request(self) -> None:
        """Send the request to fetch each voltage phase and stream 3-phase voltage."""
        # Imported here to avoid a circular import.
        from ..microgrid._data_sourcing import (  # pylint: disable=import-outside-toplevel
            ComponentMetricRequest,
        )

        metric_ids = (
            ComponentMetricId.VOLTAGE_PHASE_1,
            ComponentMetricId.VOLTAGE_PHASE_2,
            ComponentMetricId.VOLTAGE_PHASE_3,
        )
        phases_rx: list[Receiver[Sample[Quantity]]] = []
        for metric_id in metric_ids:
            req = ComponentMetricRequest(
                self._namespace, self._source_component.component_id, metric_id, None
            )

            await self._resampler_subscription_sender.send(req)

            phases_rx.append(
                self._channel_registry.get_or_create(
                    Sample[Quantity], req.get_channel_name()
                ).new_receiver()
            )

        sender = self._channel_registry.get_or_create(
            Sample3Phase[Voltage], self._channel_key
        ).new_sender()

        _logger.debug(
            "Sent request for fetching voltage from: %s", self._source_component
        )

        while True:
            try:
                phases = [await r.receive() for r in phases_rx]

                if not all(map(lambda p: p is not None, phases)):
                    _logger.warning(
                        "Received None from voltage request: %s %s",
                        self._source_component,
                        phases,
                    )
                    continue

                msg = Sample3Phase(
                    phases[0].timestamp,
                    *map(
                        lambda p: (
                            Voltage.from_volts(p.value.base_value) if p.value else None
                        ),
                        phases,
                    ),
                )
            except asyncio.CancelledError:
                _logger.exception(
                    "Phase-to-neutral 3-phase voltage streaming task cancelled: %s",
                    self._source_component,
                )
                break
            else:
                await sender.send(msg)



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Status tracking for components."""

from ._battery_status_tracker import BatteryStatusTracker
from ._component_status import (
    ComponentPoolStatus,
    ComponentStatus,
    ComponentStatusEnum,
    ComponentStatusTracker,
    SetPowerResult,
)
from ._ev_charger_status_tracker import EVChargerStatusTracker
from ._pv_inverter_status_tracker import PVInverterStatusTracker

__all__ = [
    "BatteryStatusTracker",
    "ComponentPoolStatus",
    "ComponentStatus",
    "ComponentStatusEnum",
    "ComponentStatusTracker",
    "EVChargerStatusTracker",
    "PVInverterStatusTracker",
    "SetPowerResult",
]



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""An external interface for the BatteryPool.

Allows for actors interested in operating on the same set of batteries to share
underlying formula engine and metric calculator instances, but without having to specify
their individual priorities with each request.
"""

import asyncio
import uuid
from collections import abc

from frequenz.quantities import Energy, Percentage, Power, Temperature

from ... import timeseries
from ..._internal._channels import MappingReceiverFetcher, ReceiverFetcher
from ...microgrid import _power_distributing, _power_managing
from ...timeseries import Sample
from .._base_types import SystemBounds
from ..formula_engine import FormulaEngine
from ..formula_engine._formula_generators import (
    BatteryPowerFormula,
    FormulaGeneratorConfig,
)
from ._battery_pool_reference_store import BatteryPoolReferenceStore
from ._methods import SendOnUpdate
from ._metric_calculator import (
    CapacityCalculator,
    PowerBoundsCalculator,
    SoCCalculator,
    TemperatureCalculator,
)
from .messages import BatteryPoolReport

# pylint: disable=protected-access


class BatteryPool:
    """An interface for interaction with pools of batteries.

    Provides:
      - properties for fetching reporting streams of instantaneous
        [power][frequenz.sdk.timeseries.battery_pool.BatteryPool.power],
        [soc][frequenz.sdk.timeseries.battery_pool.BatteryPool.soc],
        [capacity][frequenz.sdk.timeseries.battery_pool.BatteryPool.capacity] values and
        available power bounds and other status through
        [power_status][frequenz.sdk.timeseries.battery_pool.BatteryPool.power_status].
      - control methods for proposing power values, namely:
        [propose_power][frequenz.sdk.timeseries.battery_pool.BatteryPool.propose_power],
        [propose_charge][frequenz.sdk.timeseries.battery_pool.BatteryPool.propose_charge] and
        [propose_discharge][frequenz.sdk.timeseries.battery_pool.BatteryPool.propose_discharge].
    """

    def __init__(
        self,
        *,
        pool_ref_store: BatteryPoolReferenceStore,
        name: str | None,
        priority: int,
        set_operating_point: bool,
    ):
        """Create a BatteryPool instance.

        !!! note
            `BatteryPool` instances are not meant to be created directly by users.  Use
            the [`microgrid.new_battery_pool`][frequenz.sdk.microgrid.new_battery_pool]
            method for creating `BatteryPool` instances.

        Args:
            pool_ref_store: The battery pool reference store instance.
            name: An optional name used to identify this instance of the pool or a
                corresponding actor in the logs.
            priority: The priority of the actor using this wrapper.
            set_operating_point: Whether this instance sets the operating point power or
                the normal power for the components.
        """
        self._pool_ref_store = pool_ref_store
        unique_id = str(uuid.uuid4())
        self._source_id = unique_id if name is None else f"{name}-{unique_id}"
        self._priority = priority
        self._set_operating_point = set_operating_point

    async def propose_power(
        self,
        power: Power | None,
        *,
        bounds: timeseries.Bounds[Power | None] = timeseries.Bounds(None, None),
    ) -> None:
        """Send a proposal to the power manager for the pool's set of batteries.

        Power values need to follow the Passive Sign Convention (PSC). That is, positive
        values indicate charge power and negative values indicate discharge power.

        If the same batteries are shared by multiple actors, the power manager will
        consider the priority of the actors, the bounds they set, and their preferred
        power, when calculating the target power for the batteries.

        The preferred power of lower priority actors will take precedence as long as
        they respect the bounds set by higher priority actors.  If lower priority actors
        request power values outside of the bounds set by higher priority actors, the
        target power will be the closest value to the preferred power that is within the
        bounds.

        When there are no other actors trying to use the same batteries, the actor's
        preferred power would be set as the target power, as long as it falls within the
        system power bounds for the batteries.

        The result of the request can be accessed using the receiver returned from the
        [`power_status`][frequenz.sdk.timeseries.battery_pool.BatteryPool.power_status]
        method, which also streams the bounds that an actor should comply with, based on
        its priority.

        Args:
            power: The power to propose for the batteries in the pool.  If `None`, this
                proposal will not have any effect on the target power, unless bounds are
                specified.  If both are `None`, it is equivalent to not having a
                proposal or withdrawing a previous one.
            bounds: The power bounds for the proposal.  These bounds will apply to
                actors with a lower priority, and can be overridden by bounds from
                actors with a higher priority.  If None, the power bounds will be set
                to the maximum power of the batteries in the pool.  This is currently
                and experimental feature.
        """
        await self._pool_ref_store._power_manager_requests_sender.send(
            _power_managing.Proposal(
                source_id=self._source_id,
                preferred_power=power,
                bounds=bounds,
                component_ids=self._pool_ref_store._batteries,
                priority=self._priority,
                creation_time=asyncio.get_running_loop().time(),
                set_operating_point=self._set_operating_point,
            )
        )

    async def propose_charge(self, power: Power | None) -> None:
        """Set the given charge power for the batteries in the pool.

        Power values need to be positive values, indicating charge power.

        When using the Passive Sign Convention, the
        [`propose_power`][frequenz.sdk.timeseries.battery_pool.BatteryPool.propose_power]
        method might be more convenient.

        If the same batteries are shared by multiple actors, the behaviour is the same
        as that of the `propose_power` method.  The bounds for lower priority actors
        can't be specified with this method.  If that's required, use the
        `propose_power` method instead.

        The result of the request can be accessed using the receiver returned from the
        [`power_status`][frequenz.sdk.timeseries.battery_pool.BatteryPool.power_status]
        method.

        Args:
            power: The unsigned charge power to propose for the batteries in the pool.
                If None, the proposed power of higher priority actors will take
                precedence as the target power.

        Raises:
            ValueError: If the given power is negative.
        """
        if power and power < Power.zero():
            raise ValueError("Charge power must be positive.")
        await self._pool_ref_store._power_manager_requests_sender.send(
            _power_managing.Proposal(
                source_id=self._source_id,
                preferred_power=power,
                bounds=timeseries.Bounds(None, None),
                component_ids=self._pool_ref_store._batteries,
                priority=self._priority,
                creation_time=asyncio.get_running_loop().time(),
                set_operating_point=self._set_operating_point,
            )
        )

    async def propose_discharge(self, power: Power | None) -> None:
        """Set the given discharge power for the batteries in the pool.

        Power values need to be positive values, indicating discharge power.

        When using the Passive Sign Convention, the
        [`propose_power`][frequenz.sdk.timeseries.battery_pool.BatteryPool.propose_power]
        method might be more convenient.

        If the same batteries are shared by multiple actors, the behaviour is the same
        as that of the `propose_power` method.  The bounds for lower priority actors
        can't be specified with this method.  If that's required, use the
        `propose_power` method instead.

        The result of the request can be accessed using the receiver returned from the
        [`power_status`][frequenz.sdk.timeseries.battery_pool.BatteryPool.power_status]
        method.

        Args:
            power: The unsigned discharge power to propose for the batteries in the
                pool.  If None, the proposed power of higher priority actors will take
                precedence as the target power.

        Raises:
            ValueError: If the given power is negative.
        """
        if power:
            if power < Power.zero():
                raise ValueError("Discharge power must be positive.")
            power = -power
        await self._pool_ref_store._power_manager_requests_sender.send(
            _power_managing.Proposal(
                source_id=self._source_id,
                preferred_power=power,
                bounds=timeseries.Bounds(None, None),
                component_ids=self._pool_ref_store._batteries,
                priority=self._priority,
                creation_time=asyncio.get_running_loop().time(),
                set_operating_point=self._set_operating_point,
            )
        )

    @property
    def component_ids(self) -> abc.Set[int]:
        """Return ids of the batteries in the pool.

        Returns:
            Ids of the batteries in the pool
        """
        return self._pool_ref_store._batteries

    @property
    def power(self) -> FormulaEngine[Power]:
        """Fetch the total power of the batteries in the pool.

        This formula produces values that are in the Passive Sign Convention (PSC).

        If a formula engine to calculate this metric is not already running, it will be
        started.

        A receiver from the formula engine can be obtained by calling the `new_receiver`
        method.

        Returns:
            A FormulaEngine that will calculate and stream the total power of all
                batteries in the pool.
        """
        engine = self._pool_ref_store._formula_pool.from_power_formula_generator(
            "battery_pool_power",
            BatteryPowerFormula,
            FormulaGeneratorConfig(
                component_ids=self._pool_ref_store._batteries,
                allow_fallback=True,
            ),
        )
        assert isinstance(engine, FormulaEngine)
        return engine

    @property
    def soc(self) -> ReceiverFetcher[Sample[Percentage]]:
        """Fetch the normalized average weighted-by-capacity SoC values for the pool.

        The SoC values are normalized to the 0-100% range and clamped if they are out
        of bounds. Only values from working batteries with operational inverters are
        considered in the calculation.

        Average SoC is calculated using the formula:
        ```
        working_batteries: Set[BatteryData] # working batteries from the battery pool

        soc_scaled = min(max(
            0,
            (soc - soc_lower_bound) / (soc_upper_bound - soc_lower_bound) * 100,
        ), 100)
        used_capacity = sum(
            battery.usable_capacity * battery.soc_scaled
            for battery in working_batteries
        )
        total_capacity = sum(battery.usable_capacity for battery in working_batteries)
        average_soc = used_capacity/total_capacity
        ```

        `None` values will be sent if there are no working batteries with operational
        inverters to calculate the metric with.

        A receiver from the MetricAggregator can be obtained by calling the
        `new_receiver` method.

        Returns:
            A MetricAggregator that will calculate and stream the aggregate SoC of all
                batteries in the pool, considering only working batteries with
                operational inverters.
        """
        method_name = SendOnUpdate.name() + "_" + SoCCalculator.name()

        if method_name not in self._pool_ref_store._active_methods:
            calculator = SoCCalculator(self._pool_ref_store._batteries)
            self._pool_ref_store._active_methods[method_name] = SendOnUpdate(
                metric_calculator=calculator,
                working_batteries=self._pool_ref_store._working_batteries,
                min_update_interval=self._pool_ref_store._min_update_interval,
            )

        return self._pool_ref_store._active_methods[method_name]

    @property
    def temperature(self) -> ReceiverFetcher[Sample[Temperature]]:
        """Fetch the average temperature of the batteries in the pool.

        Returns:
            A MetricAggregator that will calculate and stream the average temperature
                of all batteries in the pool.
        """
        method_name = SendOnUpdate.name() + "_" + TemperatureCalculator.name()
        if method_name not in self._pool_ref_store._active_methods:
            calculator = TemperatureCalculator(self._pool_ref_store._batteries)
            self._pool_ref_store._active_methods[method_name] = SendOnUpdate(
                metric_calculator=calculator,
                working_batteries=self._pool_ref_store._working_batteries,
                min_update_interval=self._pool_ref_store._min_update_interval,
            )
        return self._pool_ref_store._active_methods[method_name]

    @property
    def capacity(self) -> ReceiverFetcher[Sample[Energy]]:
        """Get a receiver to receive new capacity metrics when they change.

        The reported capacity values consider only working batteries with operational
        inverters.

        Calculated with the formula:
        ```
        working_batteries: Set[BatteryData] # working batteries from the battery pool
        total_capacity = sum(
            battery.capacity * (soc_upper_bound - soc_lower_bound) / 100
            for battery in working_batteries
        )
        ```

        `None` will be sent if there are no working batteries with operational
        inverters to calculate metrics.

        A receiver from the MetricAggregator can be obtained by calling the
        `new_receiver` method.

        Returns:
            A MetricAggregator that will calculate and stream the capacity of all
                batteries in the pool, considering only working batteries with
                operational inverters.
        """
        method_name = SendOnUpdate.name() + "_" + CapacityCalculator.name()

        if method_name not in self._pool_ref_store._active_methods:
            calculator = CapacityCalculator(self._pool_ref_store._batteries)
            self._pool_ref_store._active_methods[method_name] = SendOnUpdate(
                metric_calculator=calculator,
                working_batteries=self._pool_ref_store._working_batteries,
                min_update_interval=self._pool_ref_store._min_update_interval,
            )

        return self._pool_ref_store._active_methods[method_name]

    @property
    def power_status(self) -> ReceiverFetcher[BatteryPoolReport]:
        """Get a receiver to receive new power status reports when they change.

        These include
          - the current inclusion/exclusion bounds available for the pool's priority,
          - the current target power for the pool's set of batteries,
          - the result of the last distribution request for the pool's set of batteries.

        Returns:
            A receiver that will stream power status reports for the pool's priority.
        """
        sub = _power_managing.ReportRequest(
            source_id=self._source_id,
            priority=self._priority,
            component_ids=self._pool_ref_store._batteries,
            set_operating_point=self._set_operating_point,
        )
        self._pool_ref_store._power_bounds_subs[sub.get_channel_name()] = (
            asyncio.create_task(
                self._pool_ref_store._power_manager_bounds_subscription_sender.send(sub)
            )
        )
        channel = self._pool_ref_store._channel_registry.get_or_create(
            _power_managing._Report, sub.get_channel_name()
        )
        channel.resend_latest = True

        return channel

    @property
    def power_distribution_results(self) -> ReceiverFetcher[_power_distributing.Result]:
        """Get a receiver to receive power distribution results.

        Returns:
            A receiver that will stream power distribution results for the pool's set of
            batteries.
        """
        return MappingReceiverFetcher(
            self._pool_ref_store._power_dist_results_fetcher,
            lambda recv: recv.filter(
                lambda x: x.request.component_ids == self._pool_ref_store._batteries
            ),
        )

    @property
    def _system_power_bounds(self) -> ReceiverFetcher[SystemBounds]:
        """Get receiver to receive new power bounds when they change.

        Power bounds refer to the min and max power that a battery can
        discharge or charge at and is also denoted as SoP.

        Power bounds formulas are described in the receiver return type.
        None will be send if there is no component to calculate metrics.

        A receiver from the MetricAggregator can be obtained by calling the
        `new_receiver` method.

        Returns:
            A MetricAggregator that will calculate and stream the power bounds
                of all batteries in the pool.
        """
        method_name = SendOnUpdate.name() + "_" + PowerBoundsCalculator.name()

        if method_name not in self._pool_ref_store._active_methods:
            calculator = PowerBoundsCalculator(self._pool_ref_store._batteries)
            self._pool_ref_store._active_methods[method_name] = SendOnUpdate(
                metric_calculator=calculator,
                working_batteries=self._pool_ref_store._working_batteries,
                min_update_interval=self._pool_ref_store._min_update_interval,
            )

        return self._pool_ref_store._active_methods[method_name]

    async def stop(self) -> None:
        """Stop all tasks and channels owned by the BatteryPool."""
        await self._pool_ref_store.stop()



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Provides a way for the SDK to apply formulas on resampled data streams.

# Formula Engine

[`FormulaEngine`][frequenz.sdk.timeseries.formula_engine.FormulaEngine]s are used in the
SDK to calculate and stream metrics like
[`grid_power`][frequenz.sdk.timeseries.grid.Grid.power],
[`consumer_power`][frequenz.sdk.timeseries.consumer.Consumer.power], etc., which are
building blocks of the [Frequenz SDK Microgrid
Model][frequenz.sdk.microgrid--frequenz-sdk-microgrid-model].

The SDK creates the formulas by analysing the configuration of components in the
{{glossary("Component Graph")}}.

## Streaming Interface

The
[`FormulaEngine.new_receiver()`][frequenz.sdk.timeseries.formula_engine.FormulaEngine.new_receiver]
method can be used to create a [Receiver][frequenz.channels.Receiver] that streams the
[Sample][frequenz.sdk.timeseries.Sample]s calculated by the formula engine.

```python
from frequenz.sdk import microgrid

battery_pool = microgrid.new_battery_pool(priority=5)

async for power in battery_pool.power.new_receiver():
    print(f"{power=}")
```

## Composition

Composite `FormulaEngine`s can be built using arithmetic operations on
`FormulaEngine`s streaming the same type of data.

For example, if you're interested in a particular composite metric that can be
calculated by subtracting
[`new_battery_pool().power`][frequenz.sdk.timeseries.battery_pool.BatteryPool.power] and
[`new_ev_charger_pool().power`][frequenz.sdk.timeseries.ev_charger_pool.EVChargerPool]
from the
[`grid().power`][frequenz.sdk.timeseries.grid.Grid.power],
we can build a `FormulaEngine` that provides a stream of this calculated metric as
follows:

```python
from frequenz.sdk import microgrid

battery_pool = microgrid.new_battery_pool(priority=5)
ev_charger_pool = microgrid.new_ev_charger_pool(priority=5)
grid = microgrid.grid()

# apply operations on formula engines to create a formula engine that would
# apply these operations on the corresponding data streams.
net_power = (
    grid.power - (battery_pool.power + ev_charger_pool.power)
).build("net_power")

async for power in net_power.new_receiver():
    print(f"{power=}")
```

# Formula Engine 3-Phase

A [`FormulaEngine3Phase`][frequenz.sdk.timeseries.formula_engine.FormulaEngine3Phase]
is similar to a
[`FormulaEngine`][frequenz.sdk.timeseries.formula_engine.FormulaEngine], except that
they stream [3-phase samples][frequenz.sdk.timeseries.Sample3Phase].  All the
current formulas (like
[`Grid.current_per_phase`][frequenz.sdk.timeseries.grid.Grid.current_per_phase],
[`EVChargerPool.current_per_phase`][frequenz.sdk.timeseries.ev_charger_pool.EVChargerPool.current_per_phase],
etc.) are implemented as per-phase formulas.

## Streaming Interface

The
[`FormulaEngine3Phase.new_receiver()`][frequenz.sdk.timeseries.formula_engine.FormulaEngine3Phase.new_receiver]
method can be used to create a [Receiver][frequenz.channels.Receiver] that streams the
[Sample3Phase][frequenz.sdk.timeseries.Sample3Phase] values
calculated by the formula engine.

```python
from frequenz.sdk import microgrid

ev_charger_pool = microgrid.new_ev_charger_pool(priority=5)

async for sample in ev_charger_pool.current_per_phase.new_receiver():
    print(f"Current: {sample}")
```

## Composition

`FormulaEngine3Phase` instances can be composed together, just like `FormulaEngine`
instances.

```python
from frequenz.sdk import microgrid

ev_charger_pool = microgrid.new_ev_charger_pool(priority=5)
grid = microgrid.grid()

# Calculate grid consumption current that's not used by the EV chargers
other_current = (grid.current_per_phase - ev_charger_pool.current_per_phase).build(
    "other_current"
)

async for sample in other_current.new_receiver():
    print(f"Other current: {sample}")
```
"""

from ._formula_engine import FormulaEngine, FormulaEngine3Phase

__all__ = [
    "FormulaEngine",
    "FormulaEngine3Phase",
]



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Background service that tracks the status of a battery.

A battery is consider to be WORKING if both the battery and the adjacent inverter are
sending data that shows that they are working.

If either of them stops sending data, or if the data shows that they are not working,
then the battery is considered to be NOT_WORKING.

If a battery and its adjacent inverter are WORKING, but the last request to the battery
failed, then the battery's status is considered to be UNCERTAIN. In this case, the
battery is blocked for a short time, and it is not recommended to use it unless it is
necessary.
"""

import asyncio
import logging
import math
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone

from frequenz.channels import Receiver, Sender, select, selected_from
from frequenz.channels.timer import SkipMissedAndDrift, Timer
from frequenz.client.microgrid import (
    BatteryComponentState,
    BatteryData,
    BatteryRelayState,
    ComponentCategory,
    ComponentData,
    ErrorLevel,
    InverterComponentState,
    InverterData,
)
from typing_extensions import override

from ....actor._background_service import BackgroundService
from ... import connection_manager
from ._blocking_status import BlockingStatus
from ._component_status import (
    ComponentStatus,
    ComponentStatusEnum,
    ComponentStatusTracker,
    SetPowerResult,
)

_logger = logging.getLogger(__name__)


@dataclass
class _ComponentStreamStatus:
    component_id: int
    """Component id."""

    data_recv_timer: Timer
    """Timer that is set when no component data has been received for some time."""

    last_msg_timestamp: datetime = datetime.now(tz=timezone.utc)
    """Timestamp of the last message from the component."""

    last_msg_correct: bool = False
    """Flag whether last message was correct or not."""


class BatteryStatusTracker(ComponentStatusTracker, BackgroundService):
    """Class for tracking if battery is working.

    Status updates are sent out only when there is a status change.
    """

    _battery_valid_relay: set[BatteryRelayState] = {BatteryRelayState.CLOSED}
    """The list of valid relay states of a battery.

    A working battery in any other battery relay state will be reported as failing.
    """

    _battery_valid_state: set[BatteryComponentState] = {
        BatteryComponentState.IDLE,
        BatteryComponentState.CHARGING,
        BatteryComponentState.DISCHARGING,
    }
    """The list of valid states of a battery.

    A working battery in any other battery state will be reported as failing.
    """

    _inverter_valid_state: set[InverterComponentState] = {
        InverterComponentState.STANDBY,
        InverterComponentState.IDLE,
        InverterComponentState.CHARGING,
        InverterComponentState.DISCHARGING,
    }
    """The list of valid states of an inverter.

    A working inverter in any other inverter state will be reported as failing.
    """

    @override
    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        component_id: int,
        max_data_age: timedelta,
        max_blocking_duration: timedelta,
        status_sender: Sender[ComponentStatus],
        set_power_result_receiver: Receiver[SetPowerResult],
    ) -> None:
        """Create class instance.

        Args:
            component_id: Id of this battery
            max_data_age: If component stopped sending data, then this is the maximum
                time when its last message should be considered as valid. After that
                time, component won't be used until it starts sending data.
            max_blocking_duration: This value tell what should be the maximum
                timeout used for blocking failing component.
            status_sender: Channel to send status updates.
            set_power_result_receiver: Channel to receive results of the requests to the
                components.

        Raises:
            RuntimeError: If battery has no adjacent inverter.
        """
        BackgroundService.__init__(self, name=f"BatteryStatusTracker({component_id})")
        self._max_data_age = max_data_age
        self._status_sender = status_sender
        self._set_power_result_receiver = set_power_result_receiver

        # First battery is considered as not working.
        # Change status after first messages are received.
        self._last_status: ComponentStatusEnum = ComponentStatusEnum.NOT_WORKING
        self._blocking_status: BlockingStatus = BlockingStatus(
            min_duration=timedelta(seconds=1.0), max_duration=max_blocking_duration
        )
        self._timedelta_zero = timedelta(seconds=0.0)

        inverter_id = self._find_adjacent_inverter_id(component_id)
        if inverter_id is None:
            raise RuntimeError(
                f"Can't find inverter adjacent to battery: {component_id}"
            )

        self._battery: _ComponentStreamStatus = _ComponentStreamStatus(
            component_id,
            data_recv_timer=Timer(max_data_age, SkipMissedAndDrift()),
        )
        self._inverter: _ComponentStreamStatus = _ComponentStreamStatus(
            inverter_id,
            data_recv_timer=Timer(max_data_age, SkipMissedAndDrift()),
        )

        # Select needs receivers that can be get in async way only.

    @override
    def start(self) -> None:
        """Start the BatteryStatusTracker instance."""
        self._tasks.add(
            asyncio.create_task(
                self._run(self._status_sender, self._set_power_result_receiver)
            )
        )

    @property
    def battery_id(self) -> int:
        """Get battery id.

        Returns:
            Battery id
        """
        return self._battery.component_id

    def _handle_status_battery(self, bat_data: BatteryData) -> None:
        self._battery.last_msg_correct = (
            self._is_message_reliable(bat_data)
            and self._is_battery_state_correct(bat_data)
            and self._no_critical_error(bat_data)
            and self._is_capacity_present(bat_data)
        )
        self._battery.last_msg_timestamp = bat_data.timestamp
        self._battery.data_recv_timer.reset()

    def _handle_status_inverter(self, inv_data: InverterData) -> None:
        self._inverter.last_msg_correct = (
            self._is_message_reliable(inv_data)
            and self._is_inverter_state_correct(inv_data)
            and self._no_critical_error(inv_data)
        )
        self._inverter.last_msg_timestamp = inv_data.timestamp
        self._inverter.data_recv_timer.reset()

    def _handle_status_set_power_result(self, result: SetPowerResult) -> None:
        if self.battery_id in result.succeeded:
            self._blocking_status.unblock()

        elif (
            self.battery_id in result.failed
            and self._last_status != ComponentStatusEnum.NOT_WORKING
        ):
            duration = self._blocking_status.block()

            if duration > self._timedelta_zero:
                _logger.warning(
                    "battery %d failed last response. block it for %s",
                    self.battery_id,
                    duration,
                )

    def _handle_status_battery_timer(self) -> None:
        if self._battery.last_msg_correct:
            self._battery.last_msg_correct = False
            _logger.warning(
                "Battery %d stopped sending data, last timestamp: %s",
                self._battery.component_id,
                self._battery.last_msg_timestamp,
            )

    def _handle_status_inverter_timer(self) -> None:
        if self._inverter.last_msg_correct:
            self._inverter.last_msg_correct = False
            _logger.warning(
                "Inverter %d stopped sending data, last timestamp: %s",
                self._inverter.component_id,
                self._inverter.last_msg_timestamp,
            )

    def _get_new_status_if_changed(self) -> ComponentStatusEnum | None:
        current_status = self._get_current_status()
        if self._last_status != current_status:
            self._last_status = current_status
            _logger.info(
                "battery %d changed status %s",
                self.battery_id,
                str(self._last_status),
            )
            return current_status
        return None

    async def _run(
        self,
        status_sender: Sender[ComponentStatus],
        set_power_result_receiver: Receiver[SetPowerResult],
    ) -> None:
        """Process data from the components and set_power_result_receiver.

        New status is send only when it change.

        Args:
            status_sender: Channel to send status updates.
            set_power_result_receiver: Channel to receive results of the requests to the
                components.
        """
        api_client = connection_manager.get().api_client

        battery_receiver = await api_client.battery_data(self._battery.component_id)
        inverter_receiver = await api_client.inverter_data(self._inverter.component_id)

        battery = battery_receiver
        battery_timer = self._battery.data_recv_timer
        inverter_timer = self._inverter.data_recv_timer
        inverter = inverter_receiver
        set_power_result = set_power_result_receiver

        while True:
            try:
                async for selected in select(
                    battery,
                    battery_timer,
                    inverter_timer,
                    inverter,
                    set_power_result,
                ):
                    new_status = None

                    if selected_from(selected, battery):
                        self._handle_status_battery(selected.message)

                    elif selected_from(selected, inverter):
                        self._handle_status_inverter(selected.message)

                    elif selected_from(selected, set_power_result):
                        self._handle_status_set_power_result(selected.message)

                    elif selected_from(selected, battery_timer):
                        if (
                            datetime.now(tz=timezone.utc)
                            - self._battery.last_msg_timestamp
                        ) < self._max_data_age:
                            # This means that we have received data from the battery
                            # since the timer triggered, but the timer event arrived
                            # late, so we can ignore it.
                            continue
                        self._handle_status_battery_timer()

                    elif selected_from(selected, inverter_timer):
                        if (
                            datetime.now(tz=timezone.utc)
                            - self._inverter.last_msg_timestamp
                        ) < self._max_data_age:
                            # This means that we have received data from the inverter
                            # since the timer triggered, but the timer event arrived
                            # late, so we can ignore it.
                            continue
                        self._handle_status_inverter_timer()

                    else:
                        _logger.error("Unknown message returned from select")

                    new_status = self._get_new_status_if_changed()

                    if new_status is not None:
                        await status_sender.send(
                            ComponentStatus(self.battery_id, new_status)
                        )

            except Exception as err:  # pylint: disable=broad-except
                _logger.exception("BatteryStatusTracker crashed with error: %s", err)

    def _get_current_status(self) -> ComponentStatusEnum:
        """Get current battery status.

        Returns:
            Battery status.
        """
        is_msg_correct = (
            self._battery.last_msg_correct and self._inverter.last_msg_correct
        )

        if not is_msg_correct:
            return ComponentStatusEnum.NOT_WORKING
        if self._last_status == ComponentStatusEnum.NOT_WORKING:
            # If message just become correct, then try to use it
            self._blocking_status.unblock()
            return ComponentStatusEnum.WORKING
        if self._blocking_status.is_blocked():
            return ComponentStatusEnum.UNCERTAIN

        return ComponentStatusEnum.WORKING

    def _is_capacity_present(self, msg: BatteryData) -> bool:
        """Check whether the battery capacity is NaN or not.

        If battery capacity is missing, then we can't work with it.

        Args:
            msg: battery message

        Returns:
            True if battery capacity is present, false otherwise.
        """
        if math.isnan(msg.capacity):
            if self._last_status == ComponentStatusEnum.WORKING:
                _logger.warning(
                    "Battery %d capacity is NaN",
                    msg.component_id,
                )
            return False
        return True

    def _no_critical_error(self, msg: BatteryData | InverterData) -> bool:
        """Check if battery or inverter message has any critical error.

        Args:
            msg: message.

        Returns:
            True if message has no critical error, False otherwise.
        """
        critical = ErrorLevel.CRITICAL
        critical_err = next((err for err in msg.errors if err.level == critical), None)
        if critical_err is not None:
            last_status = self._last_status  # pylint: disable=protected-access
            if last_status == ComponentStatusEnum.WORKING:
                _logger.warning(
                    "Component %d has critical error: %s",
                    msg.component_id,
                    str(critical_err),
                )
            return False
        return True

    def _is_inverter_state_correct(self, msg: InverterData) -> bool:
        """Check if inverter is in correct state from message.

        Args:
            msg: message

        Returns:
            True if inverter is in correct state. False otherwise.
        """
        # Component state is not exposed to the user.
        # pylint: disable=protected-access
        state = msg.component_state
        if state not in BatteryStatusTracker._inverter_valid_state:
            if self._last_status == ComponentStatusEnum.WORKING:
                _logger.warning(
                    "Inverter %d has invalid state: %s",
                    msg.component_id,
                    state.name,
                )
            return False
        # pylint: enable=protected-access
        return True

    def _is_battery_state_correct(self, msg: BatteryData) -> bool:
        """Check if battery is in correct state from message.

        Args:
            msg: message

        Returns:
            True if battery is in correct state. False otherwise.
        """
        # Component state is not exposed to the user.
        # pylint: disable=protected-access
        state = msg.component_state
        if state not in BatteryStatusTracker._battery_valid_state:
            if self._last_status == ComponentStatusEnum.WORKING:
                _logger.warning(
                    "Battery %d has invalid state: %s",
                    self.battery_id,
                    state.name,
                )
            return False

        # Component state is not exposed to the user.
        relay_state = msg.relay_state
        if relay_state not in BatteryStatusTracker._battery_valid_relay:
            if self._last_status == ComponentStatusEnum.WORKING:
                _logger.warning(
                    "Battery %d has invalid relay state: %s",
                    self.battery_id,
                    relay_state.name,
                )
            return False
        return True
        # pylint: enable=protected-access

    def _is_timestamp_outdated(self, timestamp: datetime) -> bool:
        """Return if timestamp is to old.

        Args:
            timestamp: timestamp

        Returns:
            _True if timestamp is to old, False otherwise
        """
        now = datetime.now(tz=timezone.utc)
        diff = now - timestamp
        return diff > self._max_data_age

    def _is_message_reliable(self, message: ComponentData) -> bool:
        """Check if message is too old to be considered as reliable.

        Args:
            message: message to check

        Returns:
            True if message is reliable, False otherwise.
        """
        is_outdated = self._is_timestamp_outdated(message.timestamp)

        if is_outdated and self._last_status == ComponentStatusEnum.WORKING:
            _logger.warning(
                "Component %d stopped sending data. Last timestamp: %s.",
                message.component_id,
                str(message.timestamp),
            )

        return not is_outdated

    def _find_adjacent_inverter_id(self, battery_id: int) -> int | None:
        """Find inverter adjacent to this battery.

        Args:
            battery_id: battery id adjacent to the wanted inverter

        Returns:
            Id of the inverter. If battery hasn't adjacent inverter, then return None.
        """
        graph = connection_manager.get().component_graph
        return next(
            (
                comp.component_id
                for comp in graph.predecessors(battery_id)
                if comp.category == ComponentCategory.INVERTER
            ),
            None,
        )



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Types for exposing battery pool reports."""

import abc
import typing

from frequenz.quantities import Power

from ...microgrid._power_distributing import (
    Error,
    OutOfBounds,
    PartialFailure,
    Result,
    Success,
)
from .._base_types import Bounds


# This class is used to expose the generic reports from the PowerManager with specific
# documentation for the battery pool.
class BatteryPoolReport(typing.Protocol):
    """A status report for a battery pool."""

    @property
    def target_power(self) -> Power | None:
        """The currently set power for the batteries."""

    @property
    def bounds(self) -> Bounds[Power] | None:
        """The usable bounds for the batteries.

        These bounds are adjusted to any restrictions placed by actors with higher
        priorities.

        There might be exclusion zones within these bounds. If necessary, the
        [`adjust_to_bounds`][frequenz.sdk.timeseries.battery_pool.messages.BatteryPoolReport.adjust_to_bounds]
        method may be used to check if a desired power value fits the bounds, or to get
        the closest possible power values that do fit the bounds.
        """

    @abc.abstractmethod
    def adjust_to_bounds(self, power: Power) -> tuple[Power | None, Power | None]:
        """Adjust a power value to the bounds.

        This method can be used to adjust a desired power value to the power bounds
        available to the actor.

        If the given power value falls within the usable bounds, it will be returned
        unchanged.

        If it falls outside the usable bounds, the closest possible value on the
        corresponding side will be returned.  For example, if the given power is lower
        than the lowest usable power, only the lowest usable power will be returned, and
        similarly for the highest usable power.

        If the given power falls within an exclusion zone that's contained within the
        usable bounds, the closest possible power values on both sides will be returned.

        !!! note
            It is completely optional to use this method to adjust power values before
            proposing them, because the battery pool will do this automatically.  This
            method is provided for convenience, and for granular control when there are
            two possible power values, both of which fall within the available bounds.

        Example:
            ```python
            from frequenz.sdk import microgrid

            power_status_rx = microgrid.new_battery_pool(
                priority=5,
            ).power_status.new_receiver()
            power_status = await power_status_rx.receive()
            desired_power = Power.from_watts(1000.0)

            match power_status.adjust_to_bounds(desired_power):
                case (power, _) if power == desired_power:
                    print("Desired power is available.")
                case (None, power) | (power, None) if power:
                    print(f"Closest available power is {power}.")
                case (lower, upper) if lower and upper:
                    print(f"Two options {lower}, {upper} to propose to battery pool.")
                case (None, None):
                    print("No available power")
            ```

        Args:
            power: The power value to adjust.

        Returns:
            A tuple of the closest power values to the desired power that fall within
                the available bounds for the actor.
        """


__all__ = [
    "BatteryPoolReport",
    "Error",
    "OutOfBounds",
    "PartialFailure",
    "Result",
    "Success",
]



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Manage ev chargers for the power distributor."""

from ._ev_charger_manager import EVChargerManager

__all__ = [
    "EVChargerManager",
]



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Manage batteries and inverters for the power distributor."""

import asyncio
import collections.abc
import logging
import math
import typing
from datetime import timedelta

from frequenz.channels import LatestValueCache, Receiver, Sender
from frequenz.client.microgrid import (
    ApiClientError,
    BatteryData,
    ComponentCategory,
    InverterData,
    OperationOutOfRange,
)
from frequenz.quantities import Power
from typing_extensions import override

from ...._internal._math import is_close_to_zero
from ... import connection_manager
from .._component_pool_status_tracker import ComponentPoolStatusTracker
from .._component_status import BatteryStatusTracker, ComponentPoolStatus
from .._distribution_algorithm import (
    AggregatedBatteryData,
    BatteryDistributionAlgorithm,
    DistributionResult,
    InvBatPair,
)
from ..request import Request
from ..result import Error, OutOfBounds, PartialFailure, PowerBounds, Result, Success
from ._component_manager import ComponentManager

_logger = logging.getLogger(__name__)


def _get_all_from_map(
    source: dict[int, frozenset[int]], keys: collections.abc.Set[int]
) -> set[int]:
    """Get all values for the given keys from the given map.

    Args:
        source: map to get values from.
        keys: keys to get values for.

    Returns:
        Set of values for the given keys.
    """
    return set().union(*[source[key] for key in keys])


def _get_battery_inverter_mappings(
    battery_ids: collections.abc.Set[int],
    *,  # force keyword arguments
    inv_bats: bool = True,
    bat_bats: bool = True,
    inv_invs: bool = True,
) -> dict[str, dict[int, frozenset[int]]]:
    """Create maps between battery and adjacent inverters.

    Args:
        battery_ids: set of battery ids
        inv_bats: whether to create the inverter to batteries map
        bat_bats: whether to create the battery to batteries map
        inv_invs: whether to create the inverter to inverters map

    Returns:
        a dict of the requested maps, using the following keys:
            * "bat_invs": battery to inverters map
            * "inv_bats": inverter to batteries map
            * "bat_bats": battery to batteries map
            * "inv_invs": inverter to inverters map
    """
    bat_invs_map: dict[int, set[int]] = {}
    inv_bats_map: dict[int, set[int]] | None = {} if inv_bats else None
    bat_bats_map: dict[int, set[int]] | None = {} if bat_bats else None
    inv_invs_map: dict[int, set[int]] | None = {} if inv_invs else None
    component_graph = connection_manager.get().component_graph

    for battery_id in battery_ids:
        inverters: set[int] = set(
            component.component_id
            for component in component_graph.predecessors(battery_id)
            if component.category == ComponentCategory.INVERTER
        )

        if len(inverters) == 0:
            _logger.error("No inverters for battery %d", battery_id)
            continue

        bat_invs_map[battery_id] = inverters
        if bat_bats_map is not None:
            bat_bats_map.setdefault(battery_id, set()).update(
                set(
                    component.component_id
                    for inverter in inverters
                    for component in component_graph.successors(inverter)
                )
            )

        for inverter in inverters:
            if inv_bats_map is not None:
                inv_bats_map.setdefault(inverter, set()).add(battery_id)
            if inv_invs_map is not None:
                inv_invs_map.setdefault(inverter, set()).update(bat_invs_map)

    mapping: dict[str, dict[int, frozenset[int]]] = {}

    # Convert sets to frozensets to make them hashable.
    def _add(key: str, value: dict[int, set[int]] | None) -> None:
        if value is not None:
            mapping[key] = {k: frozenset(v) for k, v in value.items()}

    _add("bat_invs", bat_invs_map)
    _add("inv_bats", inv_bats_map)
    _add("bat_bats", bat_bats_map)
    _add("inv_invs", inv_invs_map)

    return mapping


class BatteryManager(ComponentManager):  # pylint: disable=too-many-instance-attributes
    """Class to manage the data streams for batteries."""

    @override
    def __init__(
        self,
        component_pool_status_sender: Sender[ComponentPoolStatus],
        results_sender: Sender[Result],
        api_power_request_timeout: timedelta,
    ):
        """Initialize this instance.

        Args:
            component_pool_status_sender: Channel sender to send the status of the
                battery pool to.  This status is used by the battery pool metric
                streams, to dynamically adjust the values based on the health of the
                individual batteries.
            results_sender: Channel sender to send the power distribution results to.
            api_power_request_timeout: Timeout to use when making power requests to
                the microgrid API.
        """
        self._results_sender = results_sender
        self._api_power_request_timeout = api_power_request_timeout
        self._batteries = connection_manager.get().component_graph.components(
            component_categories={ComponentCategory.BATTERY}
        )
        self._battery_ids = {battery.component_id for battery in self._batteries}

        maps = _get_battery_inverter_mappings(self._battery_ids)

        self._bat_invs_map = maps["bat_invs"]
        self._inv_bats_map = maps["inv_bats"]
        self._bat_bats_map = maps["bat_bats"]
        self._inv_invs_map = maps["inv_invs"]

        self._battery_caches: dict[int, LatestValueCache[BatteryData]] = {}
        self._inverter_caches: dict[int, LatestValueCache[InverterData]] = {}

        self._component_pool_status_tracker = ComponentPoolStatusTracker(
            component_ids=set(self._battery_ids),
            component_status_sender=component_pool_status_sender,
            max_blocking_duration=timedelta(seconds=30.0),
            max_data_age=timedelta(seconds=10.0),
            component_status_tracker_type=BatteryStatusTracker,
        )

        # NOTE: power_distributor_exponent should be received from ConfigManager
        self._power_distributor_exponent: float = 1.0
        """The exponent for the power distribution algorithm.

        The exponent determines how fast the batteries should strive to the
        equal SoC level.
        """

        self._distribution_algorithm = BatteryDistributionAlgorithm(
            self._power_distributor_exponent
        )
        """The distribution algorithm used to distribute power between batteries."""

    @override
    def component_ids(self) -> collections.abc.Set[int]:
        """Return the set of component ids."""
        return self._battery_ids

    @override
    async def start(self) -> None:
        """Start the battery data manager."""
        await self._create_channels()

    @override
    async def stop(self) -> None:
        """Stop the battery data manager."""
        for bat_cache in self._battery_caches.values():
            await bat_cache.stop()
        for inv_cache in self._inverter_caches.values():
            await inv_cache.stop()
        await self._component_pool_status_tracker.stop()

    @override
    async def distribute_power(self, request: Request) -> None:
        """Distribute the requested power to the components.

        Args:
            request: Request to get the distribution for.
        """
        distribution_result = await self._get_distribution(request)
        if not isinstance(distribution_result, DistributionResult):
            result = distribution_result
        else:
            result = await self._distribute_power(request, distribution_result)
        await self._results_sender.send(result)

    async def _get_distribution(self, request: Request) -> DistributionResult | Result:
        """Get the distribution of the batteries.

        Args:
            request: Request to get the distribution for.

        Returns:
            Distribution of the batteries.
        """
        try:
            pairs_data: list[InvBatPair] = self._get_components_data(
                request.component_ids
            )
        except KeyError as err:
            return Error(request=request, msg=str(err))

        if not pairs_data:
            error_msg = (
                "No data for at least one of the given "
                f"batteries {str(request.component_ids)}"
            )
            return Error(request=request, msg=str(error_msg))

        error = self._check_request(request, pairs_data)
        if error:
            return error

        try:
            distribution = self._get_power_distribution(request, pairs_data)
        except ValueError as err:
            _logger.exception("Couldn't distribute power")
            error_msg = f"Couldn't distribute power, error: {str(err)}"
            return Error(request=request, msg=str(error_msg))

        return distribution

    async def _distribute_power(
        self, request: Request, distribution: DistributionResult
    ) -> Result:
        """Set the distributed power to the batteries.

        Args:
            request: Request to set the power for.
            distribution: Distribution to set.

        Returns:
            Result from the microgrid API.
        """
        distributed_power_value = (
            request.power.as_watts() - distribution.remaining_power
        )
        battery_distribution: dict[int, float] = {}
        for inverter_id, dist in distribution.distribution.items():
            for battery_id in self._inv_bats_map[inverter_id]:
                battery_distribution[battery_id] = (
                    battery_distribution.get(battery_id, 0.0) + dist
                )
        _logger.debug(
            "Distributing power %d between the batteries %s",
            distributed_power_value,
            str(battery_distribution),
        )

        failed_power, failed_batteries = await self._set_distributed_power(
            distribution, self._api_power_request_timeout
        )

        response: Success | PartialFailure
        if len(failed_batteries) > 0:
            succeed_batteries = set(battery_distribution.keys()) - failed_batteries
            response = PartialFailure(
                request=request,
                succeeded_power=Power.from_watts(
                    distributed_power_value - failed_power
                ),
                succeeded_components=succeed_batteries,
                failed_power=Power.from_watts(failed_power),
                failed_components=failed_batteries,
                excess_power=Power.from_watts(distribution.remaining_power),
            )
        else:
            succeed_batteries = set(battery_distribution.keys())
            response = Success(
                request=request,
                succeeded_power=Power.from_watts(distributed_power_value),
                succeeded_components=succeed_batteries,
                excess_power=Power.from_watts(distribution.remaining_power),
            )

        await asyncio.gather(
            *[
                self._component_pool_status_tracker.update_status(
                    succeed_batteries, failed_batteries
                ),
            ]
        )

        return response

    async def _create_channels(self) -> None:
        """Create channels to get data of components in microgrid."""
        api = connection_manager.get().api_client
        manager_id = f"{type(self).__name__}«{hex(id(self))}»"
        for battery_id, inverter_ids in self._bat_invs_map.items():
            bat_recv: Receiver[BatteryData] = await api.battery_data(battery_id)
            self._battery_caches[battery_id] = LatestValueCache(
                bat_recv,
                unique_id=f"{manager_id}:battery«{battery_id}»",
            )

            for inverter_id in inverter_ids:
                inv_recv: Receiver[InverterData] = await api.inverter_data(inverter_id)
                self._inverter_caches[inverter_id] = LatestValueCache(
                    inv_recv, unique_id=f"{manager_id}:inverter«{inverter_id}»"
                )

    def _get_bounds(
        self,
        pairs_data: list[InvBatPair],
    ) -> PowerBounds:
        """Get power bounds for given batteries.

        Args:
            pairs_data: list of battery and adjacent inverter data pairs.

        Returns:
            Power bounds for given batteries.
        """
        return PowerBounds(
            inclusion_lower=sum(
                max(
                    battery.power_bounds.inclusion_lower,
                    sum(
                        inverter.active_power_inclusion_lower_bound
                        for inverter in inverters
                    ),
                )
                for battery, inverters in pairs_data
            ),
            inclusion_upper=sum(
                min(
                    battery.power_bounds.inclusion_upper,
                    sum(
                        inverter.active_power_inclusion_upper_bound
                        for inverter in inverters
                    ),
                )
                for battery, inverters in pairs_data
            ),
            exclusion_lower=min(
                sum(battery.power_bounds.exclusion_lower for battery, _ in pairs_data),
                sum(
                    inverter.active_power_exclusion_lower_bound
                    for _, inverters in pairs_data
                    for inverter in inverters
                ),
            ),
            exclusion_upper=max(
                sum(battery.power_bounds.exclusion_upper for battery, _ in pairs_data),
                sum(
                    inverter.active_power_exclusion_upper_bound
                    for _, inverters in pairs_data
                    for inverter in inverters
                ),
            ),
        )

    def _check_request(
        self,
        request: Request,
        pairs_data: list[InvBatPair],
    ) -> Result | None:
        """Check whether the given request if correct.

        Args:
            request: request to check
            pairs_data: list of battery and adjacent inverter data pairs.

        Returns:
            Result for the user if the request is wrong, None otherwise.
        """
        if not request.component_ids:
            return Error(request=request, msg="Empty battery IDs in the request")

        for battery in request.component_ids:
            _logger.debug("Checking battery %d", battery)
            if battery not in self._battery_caches:
                msg = (
                    f"No battery {battery}, available batteries: "
                    f"{list(self._battery_caches.keys())}"
                )
                return Error(request=request, msg=msg)

        bounds = self._get_bounds(pairs_data)

        power = request.power.as_watts()

        # Zero power requests are always forwarded to the microgrid API, even if they
        # are outside the exclusion bounds.
        if is_close_to_zero(power):
            return None

        if request.adjust_power:
            # Automatic power adjustments can only bring down the requested power down
            # to the inclusion bounds.
            #
            # If the requested power is in the exclusion bounds, it is NOT possible to
            # increase it so that it is outside the exclusion bounds.
            if bounds.exclusion_lower < power < bounds.exclusion_upper:
                return OutOfBounds(request=request, bounds=bounds)
        else:
            in_lower_range = bounds.inclusion_lower <= power <= bounds.exclusion_lower
            in_upper_range = bounds.exclusion_upper <= power <= bounds.inclusion_upper
            if not (in_lower_range or in_upper_range):
                return OutOfBounds(request=request, bounds=bounds)

        return None

    def _get_battery_inverter_data(
        self, battery_ids: frozenset[int], inverter_ids: frozenset[int]
    ) -> InvBatPair | None:
        """Get battery and inverter data if they are correct.

        Each float data from the microgrid can be "NaN".
        We can't do math operations on "NaN".
        So check all the metrics and if any are "NaN" then return None.

        Args:
            battery_ids: battery ids
            inverter_ids: inverter ids

        Returns:
            Data for the battery and adjacent inverter without NaN values.
                Return None if we could not replace NaN values.
        """
        # It means that nothing has been send on these channels, yet.
        # This should be handled by BatteryStatus. BatteryStatus should not return
        # this batteries as working.
        if not all(
            self._battery_caches[bat_id].has_value() for bat_id in battery_ids
        ) or not all(
            self._inverter_caches[inv_id].has_value() for inv_id in inverter_ids
        ):
            _logger.error(
                "Battery %s or inverter %s send no data, yet. They should be not used.",
                battery_ids,
                inverter_ids,
            )
            return None

        battery_data = [
            self._battery_caches[battery_id].get() for battery_id in battery_ids
        ]
        inverter_data = [
            self._inverter_caches[inverter_id].get() for inverter_id in inverter_ids
        ]

        DataType = typing.TypeVar("DataType", BatteryData, InverterData)

        def metric_is_nan(data: DataType, metrics: list[str]) -> bool:
            """Check if non-replaceable metrics are NaN."""
            assert data is not None
            return any(map(lambda metric: math.isnan(getattr(data, metric)), metrics))

        def nan_metric_in_list(data: list[DataType], metrics: list[str]) -> bool:
            """Check if any metric is NaN."""
            return any(map(lambda datum: metric_is_nan(datum, metrics), data))

        crucial_metrics_bat = [
            "soc",
            "soc_lower_bound",
            "soc_upper_bound",
            "capacity",
            "power_inclusion_lower_bound",
            "power_inclusion_upper_bound",
        ]

        crucial_metrics_inv = [
            "active_power_inclusion_lower_bound",
            "active_power_inclusion_upper_bound",
        ]

        if nan_metric_in_list(battery_data, crucial_metrics_bat):
            _logger.debug("Some metrics for battery set %s are NaN", list(battery_ids))
            return None

        if nan_metric_in_list(inverter_data, crucial_metrics_inv):
            _logger.debug(
                "Some metrics for inverter set %s are NaN", list(inverter_ids)
            )
            return None

        return InvBatPair(AggregatedBatteryData(battery_data), inverter_data)

    def _get_components_data(
        self, batteries: collections.abc.Set[int]
    ) -> list[InvBatPair]:
        """Get data for the given batteries and adjacent inverters.

        Args:
            batteries: Batteries that needs data.

        Raises:
            KeyError: If any battery in the given list doesn't exists in microgrid.

        Returns:
            Pairs of battery and adjacent inverter data.
        """
        pairs_data: list[InvBatPair] = []

        working_batteries = self._component_pool_status_tracker.get_working_components(
            batteries
        )

        for battery_id in working_batteries:
            if battery_id not in self._battery_caches:
                raise KeyError(
                    f"No battery {battery_id}, "
                    f"available batteries: {list(self._battery_caches.keys())}"
                )

        connected_inverters = _get_all_from_map(self._bat_invs_map, batteries)

        # Check to see if inverters are involved that are connected to batteries
        # that were not requested.
        batteries_from_inverters = _get_all_from_map(
            self._inv_bats_map, connected_inverters
        )

        if batteries_from_inverters != batteries:
            extra_batteries = batteries_from_inverters - batteries
            raise KeyError(
                f"Inverters {_get_all_from_map(self._bat_invs_map, extra_batteries)} "
                f"are connected to batteries that were not requested: {extra_batteries}"
            )

        # set of set of batteries one for each working_battery
        battery_sets: frozenset[frozenset[int]] = frozenset(
            self._bat_bats_map[working_battery] for working_battery in working_batteries
        )

        for battery_ids in battery_sets:
            inverter_ids: frozenset[int] = self._bat_invs_map[next(iter(battery_ids))]

            data = self._get_battery_inverter_data(battery_ids, inverter_ids)
            if data is None:
                _logger.warning(
                    "Skipping battery set %s because at least one of its messages isn't correct.",
                    list(battery_ids),
                )
                continue

            assert len(data.inverter) > 0
            pairs_data.append(data)
        return pairs_data

    def _get_power_distribution(
        self, request: Request, inv_bat_pairs: list[InvBatPair]
    ) -> DistributionResult:
        """Get power distribution result for the batteries in the request.

        Args:
            request: the power request to process.
            inv_bat_pairs: the battery and adjacent inverter data pairs.

        Returns:
            the power distribution result.
        """
        available_bat_ids = _get_all_from_map(
            self._bat_bats_map, {pair.battery.component_id for pair in inv_bat_pairs}
        )

        unavailable_bat_ids = request.component_ids - available_bat_ids
        unavailable_inv_ids: set[int] = set()

        for inverter_ids in [
            self._bat_invs_map[battery_id_set] for battery_id_set in unavailable_bat_ids
        ]:
            unavailable_inv_ids = unavailable_inv_ids.union(inverter_ids)

        result = self._distribution_algorithm.distribute_power(
            request.power.as_watts(), inv_bat_pairs
        )

        return result

    async def _set_distributed_power(
        self,
        distribution: DistributionResult,
        timeout: timedelta,
    ) -> tuple[float, set[int]]:
        """Send distributed power to the inverters.

        Args:
            distribution: Distribution result
            timeout: How long wait for the response

        Returns:
            Tuple where first element is total failed power, and the second element
            set of batteries that failed.
        """
        api = connection_manager.get().api_client

        tasks = {
            inverter_id: asyncio.create_task(api.set_power(inverter_id, power))
            for inverter_id, power in distribution.distribution.items()
        }

        _, pending = await asyncio.wait(
            tasks.values(),
            timeout=timeout.total_seconds(),
            return_when=asyncio.ALL_COMPLETED,
        )

        await self._cancel_tasks(pending)

        return self._parse_result(tasks, distribution.distribution, timeout)

    def _parse_result(
        self,
        tasks: dict[int, asyncio.Task[None]],
        distribution: dict[int, float],
        request_timeout: timedelta,
    ) -> tuple[float, set[int]]:
        """Parse the results of `set_power` requests.

        Check if any task has failed and determine the reason for failure.
        If any task did not succeed, then the corresponding battery is marked as broken.

        Args:
            tasks: A dictionary where the key is the inverter ID and the value is the task that
                set the power for this inverter. Each task should be finished or cancelled.
            distribution: A dictionary where the key is the inverter ID and the value is how much
                power was set to the corresponding inverter.
            request_timeout: The timeout that was used for the request.

        Returns:
            A tuple where the first element is the total failed power, and the second element is
            the set of batteries that failed.
        """
        failed_power: float = 0.0
        failed_batteries: set[int] = set()

        for inverter_id, aws in tasks.items():
            battery_ids = self._inv_bats_map[inverter_id]
            failed = True
            try:
                aws.result()
                failed = False
            except OperationOutOfRange as err:
                _logger.debug(
                    "Set power for battery %s failed due to out of range error: %s",
                    battery_ids,
                    err,
                )
            except ApiClientError as err:
                _logger.warning(
                    "Set power for battery %s failed, mark it as broken. Error: %s",
                    battery_ids,
                    err,
                )
            except asyncio.exceptions.CancelledError:
                _logger.warning(
                    "Battery %s didn't respond in %f sec. Mark it as broken.",
                    battery_ids,
                    request_timeout.total_seconds(),
                )
            except Exception:  # pylint: disable=broad-except
                _logger.exception(
                    "Unknown error while setting power to batteries: %s",
                    battery_ids,
                )

            if failed:
                failed_power += distribution[inverter_id]
                failed_batteries.update(battery_ids)

        return failed_power, failed_batteries

    async def _cancel_tasks(
        self, tasks: collections.abc.Iterable[asyncio.Task[typing.Any]]
    ) -> None:
        """Cancel given asyncio tasks and wait for them.

        Args:
            tasks: tasks to cancel.
        """
        for aws in tasks:
            aws.cancel()

        await asyncio.gather(*tasks, return_exceptions=True)



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""A power manager implementation that uses the matryoshka algorithm.

When there are multiple proposals from different actors for the same set of components,
the matryoshka algorithm will consider the priority of the actors, the bounds they set
and their preferred power to determine the target power for the components.

The preferred power of lower priority actors will take precedence as long as they
respect the bounds set by higher priority actors.  If lower priority actors request
power values outside the bounds set by higher priority actors, the target power will
be the closest value to the preferred power that is within the bounds.

When there is only a single proposal for a set of components, its preferred power would
be the target power, as long as it falls within the system power bounds for the
components.
"""

from __future__ import annotations

import logging
import typing
from datetime import timedelta

from frequenz.quantities import Power
from typing_extensions import override

from ... import timeseries
from . import _bounds
from ._base_classes import BaseAlgorithm, Proposal, _Report

if typing.TYPE_CHECKING:
    from ...timeseries._base_types import SystemBounds

_logger = logging.getLogger(__name__)


class Matryoshka(BaseAlgorithm):
    """The matryoshka algorithm."""

    def __init__(self, max_proposal_age: timedelta) -> None:
        """Create a new instance of the matryoshka algorithm."""
        self._max_proposal_age_sec = max_proposal_age.total_seconds()
        self._component_buckets: dict[frozenset[int], set[Proposal]] = {}
        self._target_power: dict[frozenset[int], Power] = {}

    def _calc_target_power(
        self,
        proposals: set[Proposal],
        system_bounds: SystemBounds,
    ) -> Power:
        """Calculate the target power for the given components.

        Args:
            proposals: The proposals for the given components.
            system_bounds: The system bounds for the components in the proposal.

        Returns:
            The new target power for the components.
        """
        lower_bound = (
            system_bounds.inclusion_bounds.lower
            if system_bounds.inclusion_bounds
            # if a target power exists from a previous proposal, and the system bounds
            # have become unavailable, force the target power to be zero, by narrowing
            # the bounds to zero.
            else Power.zero()
        )
        upper_bound = (
            system_bounds.inclusion_bounds.upper
            if system_bounds.inclusion_bounds
            else Power.zero()
        )

        exclusion_bounds = None
        if system_bounds.exclusion_bounds is not None and (
            system_bounds.exclusion_bounds.lower != Power.zero()
            or system_bounds.exclusion_bounds.upper != Power.zero()
        ):
            exclusion_bounds = system_bounds.exclusion_bounds

        target_power = Power.zero()
        for next_proposal in sorted(proposals, reverse=True):
            if upper_bound < lower_bound:
                break
            if next_proposal.preferred_power:
                match _bounds.clamp_to_bounds(
                    next_proposal.preferred_power,
                    lower_bound,
                    upper_bound,
                    exclusion_bounds,
                ):
                    case (None, power) | (power, None) if power:
                        target_power = power
                    case (power_low, power_high) if power_low and power_high:
                        if (
                            power_high - next_proposal.preferred_power
                            < next_proposal.preferred_power - power_low
                        ):
                            target_power = power_high
                        else:
                            target_power = power_low

            proposal_lower = next_proposal.bounds.lower or lower_bound
            proposal_upper = next_proposal.bounds.upper or upper_bound
            # If the bounds from the current proposal are fully within the exclusion
            # bounds, then don't use them to narrow the bounds further. This allows
            # subsequent proposals to not be blocked by the current proposal.
            match _bounds.check_exclusion_bounds_overlap(
                proposal_lower, proposal_upper, exclusion_bounds
            ):
                case (True, True):
                    continue
            lower_bound = max(lower_bound, proposal_lower)
            upper_bound = min(upper_bound, proposal_upper)
            lower_bound, upper_bound = _bounds.adjust_exclusion_bounds(
                lower_bound, upper_bound, exclusion_bounds
            )

        return target_power

    def _validate_component_ids(
        self,
        component_ids: frozenset[int],
        proposal: Proposal | None,
        system_bounds: SystemBounds,
    ) -> bool:
        if component_ids not in self._component_buckets:
            # if there are no previous proposals and there are no system bounds, then
            # don't calculate a target power and fail the validation.
            if (
                system_bounds.inclusion_bounds is None
                and system_bounds.exclusion_bounds is None
            ):
                if proposal is not None:
                    _logger.warning(
                        "PowerManagingActor: No system bounds available for component "
                        "IDs %s, but a proposal was given.  The proposal will be "
                        "ignored.",
                        component_ids,
                    )
                return False

            for bucket in self._component_buckets:
                if any(component_id in bucket for component_id in component_ids):
                    raise NotImplementedError(
                        f"PowerManagingActor: component IDs {component_ids} are already"
                        " part of another bucket.  Overlapping buckets are not"
                        " yet supported."
                    )
        return True

    def get_target_power(
        self,
        component_ids: frozenset[int],
    ) -> Power | None:
        """Get the target power for the given components.

        Args:
            component_ids: The component IDs to get the target power for.

        Returns:
            The target power for the given components, or `None` if there is no target
                power.
        """
        return self._target_power.get(component_ids)

    @override
    def calculate_target_power(
        self,
        component_ids: frozenset[int],
        proposal: Proposal | None,
        system_bounds: SystemBounds,
        must_return_power: bool = False,
    ) -> Power | None:
        """Calculate and return the target power for the given components.

        Args:
            component_ids: The component IDs to calculate the target power for.
            proposal: If given, the proposal to added to the bucket, before the target
                power is calculated.
            system_bounds: The system bounds for the components in the proposal.
            must_return_power: If `True`, the algorithm must return a target power,
                even if it hasn't changed since the last call.

        Returns:
            The new target power for the components, or `None` if the target power
                didn't change.

        Raises:  # noqa: DOC502
            NotImplementedError: When the proposal contains component IDs that are
                already part of another bucket.
        """
        if not self._validate_component_ids(component_ids, proposal, system_bounds):
            return None

        if proposal is not None:
            bucket = self._component_buckets.setdefault(component_ids, set())
            if proposal in bucket:
                bucket.remove(proposal)
            bucket.add(proposal)

        # If there has not been any proposal for the given components, don't calculate a
        # target power and just return `None`.
        proposals = self._component_buckets.get(component_ids)
        if proposals is None:
            return None

        target_power = self._calc_target_power(proposals, system_bounds)

        if (
            must_return_power
            or component_ids not in self._target_power
            or self._target_power[component_ids] != target_power
        ):
            self._target_power[component_ids] = target_power
            return target_power
        return None

    @override
    def get_status(
        self,
        component_ids: frozenset[int],
        priority: int,
        system_bounds: SystemBounds,
    ) -> _Report:
        """Get the bounds for the algorithm.

        Args:
            component_ids: The IDs of the components to get the bounds for.
            priority: The priority of the actor for which the bounds are requested.
            system_bounds: The system bounds for the components.

        Returns:
            The target power and the available bounds for the given components, for
                the given priority.
        """
        target_power = self._target_power.get(component_ids)
        if system_bounds.inclusion_bounds is None:
            return _Report(
                target_power=target_power,
                _inclusion_bounds=None,
                _exclusion_bounds=system_bounds.exclusion_bounds,
            )

        lower_bound = system_bounds.inclusion_bounds.lower
        upper_bound = system_bounds.inclusion_bounds.upper

        exclusion_bounds = None
        if system_bounds.exclusion_bounds is not None and (
            system_bounds.exclusion_bounds.lower != Power.zero()
            or system_bounds.exclusion_bounds.upper != Power.zero()
        ):
            exclusion_bounds = system_bounds.exclusion_bounds

        for next_proposal in sorted(
            self._component_buckets.get(component_ids, []), reverse=True
        ):
            if next_proposal.priority <= priority:
                break
            proposal_lower = next_proposal.bounds.lower or lower_bound
            proposal_upper = next_proposal.bounds.upper or upper_bound
            match _bounds.check_exclusion_bounds_overlap(
                proposal_lower, proposal_upper, exclusion_bounds
            ):
                case (True, True):
                    continue
            calc_lower_bound = max(lower_bound, proposal_lower)
            calc_upper_bound = min(upper_bound, proposal_upper)
            if calc_lower_bound <= calc_upper_bound:
                lower_bound, upper_bound = _bounds.adjust_exclusion_bounds(
                    calc_lower_bound, calc_upper_bound, exclusion_bounds
                )
            else:
                break
        return _Report(
            target_power=target_power,
            _inclusion_bounds=timeseries.Bounds[Power](
                lower=lower_bound, upper=upper_bound
            ),
            _exclusion_bounds=system_bounds.exclusion_bounds,
        )

    @override
    def drop_old_proposals(self, loop_time: float) -> None:
        """Drop old proposals.

        This will remove all proposals that have not been updated for longer than
        `max_proposal_age`.

        Args:
            loop_time: The current loop time.
        """
        for bucket in self._component_buckets.values():
            to_delete = []
            for proposal in bucket:
                if (loop_time - proposal.creation_time) > self._max_proposal_age_sec:
                    to_delete.append(proposal)
            for proposal in to_delete:
                bucket.remove(proposal)



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Class that tracks the status of pool of components."""


import asyncio
import contextlib
import logging
from collections import abc
from datetime import timedelta

from frequenz.channels import Broadcast, Merger, Receiver, Sender, merge

from ..._internal._asyncio import cancel_and_await
from ._component_status import (
    ComponentPoolStatus,
    ComponentStatus,
    ComponentStatusEnum,
    ComponentStatusTracker,
    SetPowerResult,
)

_logger = logging.getLogger(__name__)


class ComponentPoolStatusTracker:
    """Track status of components of a given category.

    Send set of working and uncertain components, when the status of any of the tracked
    components changes.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        component_ids: abc.Set[int],
        component_status_sender: Sender[ComponentPoolStatus],
        max_data_age: timedelta,
        max_blocking_duration: timedelta,
        component_status_tracker_type: type[ComponentStatusTracker],
    ) -> None:
        """Create ComponentPoolStatusTracker instance.

        Args:
            component_ids: set of component ids whose status is to be tracked.
            component_status_sender: The sender used for sending the status of the
                tracked components.
            max_data_age: If a component stops sending data, then this is the maximum
                time for which its last message should be considered as valid. After
                that time, the component won't be used until it starts sending data.
            max_blocking_duration: This value tell what should be the maximum timeout
                used for blocking failing component.
            component_status_tracker_type: component status tracker to use for tracking
                the status of the components.
        """
        self._component_ids = component_ids
        self._max_data_age = max_data_age
        self._max_blocking_duration = max_blocking_duration
        self._component_status_sender = component_status_sender
        self._component_status_tracker_type = component_status_tracker_type

        # At first no component is working, we will get notification when they start
        # working.
        self._current_status = ComponentPoolStatus(working=set(), uncertain=set())

        # Channel for sending results of requests to the components.
        self._set_power_result_channel = Broadcast[SetPowerResult](
            name="component_request_status"
        )
        self._set_power_result_sender = self._set_power_result_channel.new_sender()
        self._component_status_trackers: list[ComponentStatusTracker] = []
        self._merged_status_receiver = self._make_merged_status_receiver()

        self._task = asyncio.create_task(self._run())

    async def join(self) -> None:
        """Wait and return when the instance's task completes.

        It will not terminate the instance, which can be done with the `stop` method.
        """
        await self._task

    async def stop(self) -> None:
        """Stop the ComponentPoolStatusTracker instance."""
        await cancel_and_await(self._task)
        await self._merged_status_receiver.stop()

    def _make_merged_status_receiver(
        self,
    ) -> Merger[ComponentStatus]:
        status_receivers: list[Receiver[ComponentStatus]] = []

        for component_id in self._component_ids:
            channel: Broadcast[ComponentStatus] = Broadcast(
                name=f"component_{component_id}_status"
            )
            tracker = self._component_status_tracker_type(
                component_id=component_id,
                max_data_age=self._max_data_age,
                max_blocking_duration=self._max_blocking_duration,
                status_sender=channel.new_sender(),
                set_power_result_receiver=self._set_power_result_channel.new_receiver(),
            )
            self._component_status_trackers.append(tracker)
            status_receivers.append(channel.new_receiver())
        return merge(*status_receivers)

    async def _run(self) -> None:
        """Start tracking component status."""
        async with contextlib.AsyncExitStack() as stack:
            for tracker in self._component_status_trackers:
                await stack.enter_async_context(tracker)
            while True:
                try:
                    await self._update_status()
                except Exception as err:  # pylint: disable=broad-except
                    _logger.error(
                        "ComponentPoolStatus failed with error: %s. Restarting.", err
                    )
                    await asyncio.sleep(1.0)

    async def _update_status(self) -> None:
        async for status in self._merged_status_receiver:
            component_id = status.component_id
            if status.value == ComponentStatusEnum.WORKING:
                self._current_status.working.add(component_id)
                self._current_status.uncertain.discard(component_id)
            elif status.value == ComponentStatusEnum.UNCERTAIN:
                self._current_status.working.discard(component_id)
                self._current_status.uncertain.add(component_id)
            elif status.value == ComponentStatusEnum.NOT_WORKING:
                self._current_status.working.discard(component_id)
                self._current_status.uncertain.discard(component_id)

            await self._component_status_sender.send(self._current_status)

    async def update_status(
        self, succeeded_components: set[int], failed_components: set[int]
    ) -> None:
        """Notify which components succeeded or failed in the request.

        Components that failed will be considered as broken and will be temporarily
        blocked.

        Components that succeeded will be unblocked.

        Args:
            succeeded_components: Component IDs for which the power request succeeded.
            failed_components: Component IDs for which the power request failed.
        """
        await self._set_power_result_sender.send(
            SetPowerResult(succeeded=succeeded_components, failed=failed_components)
        )

    def get_working_components(self, components: abc.Set[int]) -> abc.Set[int]:
        """From the given set of components, return only working ones.

        Args:
            components: Set of component IDs.

        Returns:
            IDs of subset with working components.
        """
        return self._current_status.get_working_components(components)



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Formula generator from component graph for Grid Power."""

import itertools
import logging

from frequenz.client.microgrid import Component, ComponentCategory, ComponentMetricId
from frequenz.quantities import Power

from ....microgrid import connection_manager
from ...formula_engine import FormulaEngine
from ._fallback_formula_metric_fetcher import FallbackFormulaMetricFetcher
from ._formula_generator import (
    NON_EXISTING_COMPONENT_ID,
    ComponentNotFound,
    FormulaGenerationError,
    FormulaGenerator,
    FormulaGeneratorConfig,
)

_logger = logging.getLogger(__name__)


class BatteryPowerFormula(FormulaGenerator[Power]):
    """Creates a formula engine from the component graph for calculating battery power."""

    def generate(
        self,
    ) -> FormulaEngine[Power]:
        """Make a formula for the cumulative AC battery power of a microgrid.

        The calculation is performed by adding the Active Powers of all the inverters
        that are attached to batteries.

        If there's no data coming from an inverter, that inverter's power will be
        treated as 0.

        Returns:
            A formula engine that will calculate cumulative battery power values.

        Raises:
            ComponentNotFound: if there are no batteries in the component graph, or if
                they don't have an inverter as a predecessor.
            FormulaGenerationError: If a battery has a non-inverter predecessor
                in the component graph, or if not all batteries behind a set of
                inverters have been requested.
        """
        builder = self._get_builder(
            "battery-power", ComponentMetricId.ACTIVE_POWER, Power.from_watts
        )

        if not self._config.component_ids:
            _logger.warning(
                "No Battery component IDs specified. "
                "Subscribing to the resampling actor with a non-existing "
                "component id, so that `0` values are sent from the formula."
            )
            # If there are no Batteries, we have to send 0 values as the same
            # frequency as the other streams. So we subscribe with a non-existing
            # component id, just to get a `None` message at the resampling interval.
            builder.push_component_metric(
                NON_EXISTING_COMPONENT_ID, nones_are_zeros=True
            )
            return builder.build()

        component_ids = set(self._config.component_ids)
        component_graph = connection_manager.get().component_graph
        inv_bat_mapping: dict[Component, set[Component]] = {}

        for bat_id in component_ids:
            inverters = set(
                filter(
                    component_graph.is_battery_inverter,
                    component_graph.predecessors(bat_id),
                )
            )
            if len(inverters) == 0:
                raise ComponentNotFound(
                    "All batteries must have at least one inverter as a predecessor."
                    f"Battery ID {bat_id} has no inverter as a predecessor.",
                )

            for inverter in inverters:
                all_connected_batteries = component_graph.successors(
                    inverter.component_id
                )
                battery_ids = set(
                    map(lambda battery: battery.component_id, all_connected_batteries)
                )
                if not battery_ids.issubset(component_ids):
                    raise FormulaGenerationError(
                        f"Not all batteries behind inverter {inverter.component_id} "
                        f"are requested. Missing: {battery_ids - component_ids}"
                    )

                inv_bat_mapping[inverter] = all_connected_batteries

        if self._config.allow_fallback:
            fallbacks = self._get_fallback_formulas(inv_bat_mapping)

            for idx, (primary_component, fallback_formula) in enumerate(
                fallbacks.items()
            ):
                if idx > 0:
                    builder.push_oper("+")

                builder.push_component_metric(
                    primary_component.component_id,
                    nones_are_zeros=(
                        primary_component.category != ComponentCategory.METER
                    ),
                    fallback=fallback_formula,
                )
        else:
            for idx, comp in enumerate(inv_bat_mapping.keys()):
                if idx > 0:
                    builder.push_oper("+")
                builder.push_component_metric(comp.component_id, nones_are_zeros=True)

        return builder.build()

    def _get_fallback_formulas(
        self, inv_bat_mapping: dict[Component, set[Component]]
    ) -> dict[Component, FallbackFormulaMetricFetcher[Power] | None]:
        """Find primary and fallback components and create fallback formulas.

        The primary component is the one that will be used to calculate the battery power.
        If it is not available, the fallback formula will be used instead.
        Fallback formulas calculate the battery power using the fallback components.
        Fallback formulas are wrapped in `FallbackFormulaMetricFetcher`.

        Args:
            inv_bat_mapping: A mapping from inverter to connected batteries.

        Returns:
            A dictionary mapping primary components to their FallbackFormulaMetricFetcher.
        """
        fallbacks = self._get_metric_fallback_components(set(inv_bat_mapping.keys()))

        fallback_formulas: dict[
            Component, FallbackFormulaMetricFetcher[Power] | None
        ] = {}
        for primary_component, fallback_components in fallbacks.items():
            if len(fallback_components) == 0:
                fallback_formulas[primary_component] = None
                continue

            battery_ids = set(
                map(
                    lambda battery: battery.component_id,
                    itertools.chain.from_iterable(
                        inv_bat_mapping[inv] for inv in fallback_components
                    ),
                )
            )

            generator = BatteryPowerFormula(
                f"{self._namespace}_fallback_{battery_ids}",
                self._channel_registry,
                self._resampler_subscription_sender,
                FormulaGeneratorConfig(
                    component_ids=battery_ids,
                    allow_fallback=False,
                ),
            )

            fallback_formulas[primary_component] = FallbackFormulaMetricFetcher(
                generator
            )

        return fallback_formulas



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""System bounds tracker for the EV chargers."""


import asyncio
from collections import abc

from frequenz.channels import Receiver, Sender, merge, select, selected_from
from frequenz.client.microgrid import EVChargerData
from frequenz.quantities import Power

from ..._internal._asyncio import run_forever
from ...actor import BackgroundService
from ...microgrid import connection_manager
from ...microgrid._power_distributing._component_status import ComponentPoolStatus
from .._base_types import Bounds, SystemBounds


class EVCSystemBoundsTracker(BackgroundService):
    """Track the system bounds for the EV chargers.

    System bounds are the aggregate bounds for the EV chargers in the pool that are
    working and have an EV attached to them.  They are calculated from the individual
    bounds received from the microgrid API.

    The system bounds are sent to the `bounds_sender` whenever they change.
    """

    def __init__(
        self,
        component_ids: abc.Set[int],
        status_receiver: Receiver[ComponentPoolStatus],
        bounds_sender: Sender[SystemBounds],
    ):
        """Initialize this instance.

        Args:
            component_ids: The ids of the components to track.
            status_receiver: A receiver that streams the status of the EV Chargers in
                the pool.
            bounds_sender: A sender to send the system bounds to.
        """
        super().__init__()

        self._component_ids = component_ids
        self._status_receiver = status_receiver
        self._bounds_sender = bounds_sender
        self._latest_component_data: dict[int, EVChargerData] = {}
        self._last_sent_bounds: SystemBounds | None = None
        self._component_pool_status = ComponentPoolStatus(set(), set())

    def start(self) -> None:
        """Start the EV charger system bounds tracker."""
        self._tasks.add(asyncio.create_task(run_forever(self._run)))

    async def _send_bounds(self) -> None:
        """Calculate and send the aggregate system bounds if they have changed."""
        if not self._latest_component_data:
            return
        inclusion_bounds = Bounds(
            lower=Power.from_watts(
                sum(
                    data.active_power_inclusion_lower_bound
                    for data in self._latest_component_data.values()
                )
            ),
            upper=Power.from_watts(
                sum(
                    data.active_power_inclusion_upper_bound
                    for data in self._latest_component_data.values()
                )
            ),
        )
        exclusion_bounds = Bounds(
            lower=Power.from_watts(
                sum(
                    data.active_power_exclusion_lower_bound
                    for data in self._latest_component_data.values()
                )
            ),
            upper=Power.from_watts(
                sum(
                    data.active_power_exclusion_upper_bound
                    for data in self._latest_component_data.values()
                )
            ),
        )

        if (
            self._last_sent_bounds is None
            or inclusion_bounds != self._last_sent_bounds.inclusion_bounds
            or exclusion_bounds != self._last_sent_bounds.exclusion_bounds
        ):
            self._last_sent_bounds = SystemBounds(
                timestamp=max(
                    msg.timestamp for msg in self._latest_component_data.values()
                ),
                inclusion_bounds=inclusion_bounds,
                exclusion_bounds=exclusion_bounds,
            )
            await self._bounds_sender.send(self._last_sent_bounds)

    async def _run(self) -> None:
        """Run the system bounds tracker."""
        api_client = connection_manager.get().api_client
        status_rx = self._status_receiver
        ev_data_rx = merge(
            *(
                await asyncio.gather(
                    *[api_client.ev_charger_data(cid) for cid in self._component_ids]
                )
            )
        )

        async for selected in select(status_rx, ev_data_rx):
            if selected_from(selected, status_rx):
                self._component_pool_status = selected.message
                to_remove = []
                for comp_id in self._latest_component_data:
                    if (
                        comp_id not in self._component_pool_status.working
                        and comp_id not in self._component_pool_status.uncertain
                    ):
                        to_remove.append(comp_id)
                for comp_id in to_remove:
                    del self._latest_component_data[comp_id]
            elif selected_from(selected, ev_data_rx):
                data = selected.message
                comp_id = data.component_id
                if (
                    comp_id not in self._component_pool_status.working
                    and comp_id not in self._component_pool_status.uncertain
                ):
                    continue
                self._latest_component_data[data.component_id] = data

            await self._send_bounds()



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Class that stores values of the component metrics."""


from collections.abc import Mapping
from datetime import datetime

from frequenz.client.microgrid import ComponentMetricId


class ComponentMetricsData:
    """Store values of the component metrics."""

    def __init__(
        self,
        component_id: int,
        timestamp: datetime,
        metrics: Mapping[ComponentMetricId, float],
    ) -> None:
        """Create class instance.

        Args:
            component_id: component id
            timestamp: timestamp the same for all metrics
            metrics: map between metrics and its values.
        """
        self._component_id = component_id
        self._timestamp = timestamp
        self._metrics: Mapping[ComponentMetricId, float] = metrics

    @property
    def component_id(self) -> int:
        """Get component id of the given metrics.

        Returns:
            Component id
        """
        return self._component_id

    @property
    def timestamp(self) -> datetime:
        """Get timestamp of the given metrics.

        Returns:
            Timestamp (one for all metrics).
        """
        return self._timestamp

    def get(self, metric: ComponentMetricId) -> float | None:
        """Get metric value.

        Args:
            metric: metric id

        Returns:
            Value of the metric.
        """
        return self._metrics.get(metric, None)

    def __eq__(self, other: object) -> bool:
        """Compare two objects of this class.

        Object are considered as equal if all stored values except for timestamp
        are equal.

        Args:
            other: object to compare.

        Returns:
            True if two objects are equal, false otherwise.
        """
        if not isinstance(other, ComponentMetricsData):
            return False

        return (
            self.component_id == other.component_id and self._metrics == other._metrics
        )



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""A post-fix formula evaluator that operates on `Sample` receivers."""

import asyncio
from collections.abc import Callable
from datetime import datetime
from math import isinf, isnan
from typing import Generic

from .._base_types import QuantityT, Sample
from ._formula_steps import FormulaStep, MetricFetcher


class FormulaEvaluator(Generic[QuantityT]):
    """A post-fix formula evaluator that operates on `Sample` receivers."""

    def __init__(
        self,
        name: str,
        steps: list[FormulaStep],
        metric_fetchers: dict[str, MetricFetcher[QuantityT]],
        create_method: Callable[[float], QuantityT],
    ) -> None:
        """Create a `FormulaEngine` instance.

        Args:
            name: A name for the formula.
            steps: Steps for the engine to execute, in post-fix order.
            metric_fetchers: Fetchers for each metric stream the formula depends on.
            create_method: A method to generate the output `Sample` value with.  If the
                formula is for generating power values, this would be
                `Power.from_watts`, for example.
        """
        self._name = name
        self._steps = steps
        self._metric_fetchers: dict[str, MetricFetcher[QuantityT]] = metric_fetchers
        self._first_run = True
        self._create_method: Callable[[float], QuantityT] = create_method

    async def _synchronize_metric_timestamps(
        self, metrics: set[asyncio.Task[Sample[QuantityT] | None]]
    ) -> datetime:
        """Synchronize the metric streams.

        For synchronised streams like data from the `ComponentMetricsResamplingActor`,
        this a call to this function is required only once, before the first set of
        inputs are fetched.

        Args:
            metrics: The finished tasks from the first `fetch_next` calls to all the
                `MetricFetcher`s.

        Returns:
            The timestamp of the latest metric value.

        Raises:
            RuntimeError: when some streams have no value, or when the synchronization
                of timestamps fails.
        """
        metrics_by_ts: dict[datetime, list[str]] = {}
        for metric in metrics:
            result = metric.result()
            name = metric.get_name()
            if result is None:
                raise RuntimeError(f"Stream closed for component: {name}")
            metrics_by_ts.setdefault(result.timestamp, []).append(name)
        latest_ts = max(metrics_by_ts)

        # fetch the metrics with non-latest timestamps again until we have the values
        # for the same ts for all metrics.
        for metric_ts, names in metrics_by_ts.items():
            if metric_ts == latest_ts:
                continue
            while metric_ts < latest_ts:
                for name in names:
                    fetcher = self._metric_fetchers[name]
                    next_val = await fetcher.fetch_next()
                    assert next_val is not None
                    metric_ts = next_val.timestamp
            if metric_ts > latest_ts:
                raise RuntimeError(
                    "Unable to synchronize resampled metric timestamps, "
                    f"for formula: {self._name}"
                )
        self._first_run = False
        return latest_ts

    async def apply(self) -> Sample[QuantityT]:
        """Fetch the latest metrics, apply the formula once and return the result.

        Returns:
            The result of the formula.

        Raises:
            RuntimeError: if some samples didn't arrive, or if formula application
                failed.
        """
        eval_stack: list[float] = []
        ready_metrics, pending = await asyncio.wait(
            [
                asyncio.create_task(fetcher.fetch_next(), name=name)
                for name, fetcher in self._metric_fetchers.items()
            ],
            return_when=asyncio.ALL_COMPLETED,
        )

        if pending or any(res.result() is None for res in iter(ready_metrics)):
            raise RuntimeError(
                f"Some resampled metrics didn't arrive, for formula: {self._name}"
            )

        if self._first_run:
            metric_ts = await self._synchronize_metric_timestamps(ready_metrics)
        else:
            sample = next(iter(ready_metrics)).result()
            assert sample is not None
            metric_ts = sample.timestamp

        for step in self._steps:
            step.apply(eval_stack)

        # if all steps were applied and the formula was correct, there should only be a
        # single value in the evaluation stack, and that would be the formula result.
        if len(eval_stack) != 1:
            raise RuntimeError(f"Formula application failed: {self._name}")

        res = eval_stack.pop()
        if isnan(res) or isinf(res):
            return Sample(metric_ts, None)

        return Sample(metric_ts, self._create_method(res))
